# -*- coding: utf-8 -*-
"""
Created on Wed Apr  5 09:45:53 2023

@author: Bart Steemans. Govers Lab.
"""
from .features import Features
from . import utilities as u

from tqdm import tqdm
import numpy as np
import pandas as pd  
import warnings
warnings.filterwarnings("ignore")

class Image:
    
    
    def __init__(self, image, image_name, 
                 frame, mask = None, 
                 processed_mesh_dataframe = None, 
                 px = 0.065):
        
        
        self.mask = mask
        self.cells = []
        self.px = px
        self.image_name = image_name
        self.frame = frame
        self.image = image
        self.channel = None
        
        self.joined_mask = None
        self.mesh_dataframe = None
        self.processed_mesh_dataframe = processed_mesh_dataframe
        
        self.im_interp2d = None
        self.inverted_image = None
        
        
    def calculate_features(self, add_profiling_data=True, svm=False):
        
        self.features = Features(self)
        result_df = None
    
        for cell in self.cells:
            if svm:
                self.features.get_svm_features(cell)
            else:
                self.features.get_all_features(cell, add_profiling_data)
    
        if svm:
            result_df = self.svm_features_to_dataframe()
        else:
            result_df = self.features_to_dataframe()
    
        self.im_interp2d = None
        self.inverted_image = None
        del self.features
    
        return result_df
    
    def features_to_dataframe(self):
        
        contour_df = self.dataframe_from_dict("contour_features")
        
        phaco_df = self.dataframe_from_dict("phaco_features")
        
        nucleoid_df = self.dataframe_from_dict("nucleoid_features")
        
        # Combine all feature dictionaries in 
        features_df = pd.concat([contour_df, phaco_df, nucleoid_df], axis=1)
        
        self.add_info_feature_df(features_df)
        
        return features_df
    
    def svm_features_to_dataframe(self):

        svm_df = self.dataframe_from_dict("svm_features")

        self.add_info_feature_df(svm_df)
        
        return svm_df    
    
    def dataframe_from_dict(self, feature_dictionary):
        
        data = {cell.cell_id: getattr(cell, feature_dictionary)
                for cell in self.cells if not cell.discard}
        
        df = pd.DataFrame.from_dict(data, orient='index')
        
        return df
    

    
    def add_info_feature_df(self, dataframe):
        
        dataframe.insert(0, 'image_name', self.image_name)
        dataframe.insert(0, 'cell_id', [cell.cell_id for cell in self.cells if not cell.discard])
        dataframe.insert(0, 'frame', self.frame)
        
    def join_split_pipeline(self, join_thresh = 4, split_thresh = 0.3):
        
        """
        Method used in the ic.batch_mesh_process() method.
        Returns 'processed_mesh_dataframe' which is used to created Cell objects.

        """
        self.join_cells(join_thresh)
        self.mask2mesh()
        self.split_cells(split_thresh)
        self.create_cell_object()
            
    def create_cell_object(self, verbose=True):
        self.cells = []
    
        if self.processed_mesh_dataframe is None:
            try:
                dataframe = self.mesh_dataframe
                message = 'Cell objects are created from unprocessed meshes'
    
            except AttributeError:
                print('No meshes from the masks. Either process the masks or load in mesh data from a .pkl file.')
                return
    
        else:
            dataframe = self.processed_mesh_dataframe
            message = 'Cell objects are created from processed meshes'
    
        for i, row in dataframe.iterrows():
            cell_ID = i
            contours = np.array(row['contour'])
            meshes = np.array(row['mesh'])
            midlines = np.array(row['midline'])
            cell = Cell(contours, meshes, midlines, self.image.shape, cell_ID)
            self.add_cell(cell)
    
        if verbose:
            print(message)
    
          
    def add_cell(self, cell):
        
        self.cells.append(cell)
        
    def get_inverted_image(self):
        
        inverted_image = np.subtract(np.max(self.image),self.image)
        bgr = u.phase_background(inverted_image,self.mask) #
        self.inverted_image = np.maximum(inverted_image - bgr, 0)
        self.im_interp2d = u.interp2d(self.inverted_image)
            
    def join_cells(self, thresh = None):
        
        print('Joining cells ...')    
        
        if self.mask is None:
            raise ValueError('Mask image is not loaded')
        
        pole_1 = []
        pole_2 = []
        cell_ID = []
        
        #loop through all cells in the frame
        for cell in tqdm(range(1,np.max(self.mask))): #50
            try:
                contour = u.get_contour(cell, self.mask, smoothing = 13)
                skeleton = u.extract_skeleton(contour, self.mask)
            except Exception:
                
                skeleton = []
                cell_ID.append(cell)
                continue
            
            if np.any(skeleton):
                extended_skeleton, pole1, pole2  = u.extend_skeleton(skeleton,
                                                            contour,
                                                            find_pole1=True,
                                                            find_pole2=True)
                pole_1.append(pole1)
                pole_2.append(pole2)
            else:
                cell_ID.append(cell)   
        pole_1 = np.vstack(np.array(pole_1))
        pole_2 = np.vstack(np.array(pole_2))
        
        cell_pairs = u.get_cell_pairs(pole_1, pole_2, cell_ID, thresh)
        
        if not np.any(cell_pairs):
            self.joined_mask = np.copy(self.mask)
        else:
            # Create a copy of the original masks array to store the new masks
            new_masks = np.copy(self.mask)
            dsu = {}
            for mask in np.unique(cell_pairs):
                dsu[mask + 1] = mask + 1
            # Merge masks that belong to the same group
            for mask1, mask2 in cell_pairs:
                root1 = mask1 + 1
                while root1 != dsu[root1]:
                    root1 = dsu[root1]
                root2 = mask2 + 1
                while root2 != dsu[root2]:
                    root2 = dsu[root2]
                if root1 != root2:
                    dsu[root1] = root2
    
            # Update the masks with the merged values
            for mask, root in dsu.items():
                new_masks[new_masks == mask] = root
            unique_masks = np.unique(new_masks)
            # Create a dictionary that maps unique mask values to new mask values starting from 1
            mask_dict = {mask: i for i, mask in enumerate(unique_masks)}
    
            # Use the dictionary to relabel the new_masks array and the new_labels array
            new_masks_relabel = np.vectorize(mask_dict.get)(new_masks)
            self.joined_mask = new_masks_relabel
        
    
    def mask2mesh(self):
        
        contour_fit = []
        mesh = []
        midlines = []
        
        if self.joined_mask is not None:
            mask = self.joined_mask
            print('Creating meshes from joined masks ...')
        else:
            mask = self.mask
            print('Creating meshes from raw masks ...')
        #loop through all cells in the frame
        for cell in tqdm(range(1,np.max(mask))):
            #contour and skeleton generation
            try:
                contour = u.get_contour(cell, mask, smoothing = 13)
                skeleton = u.extract_skeleton(contour, mask)
            except Exception:
                skeleton = []
                
            if np.any(skeleton):
                
                extended_skeleton, pole1, pole2  = u.extend_skeleton(skeleton,
                                                                        contour,
                                                                        find_pole1=True,
                                                                        find_pole2=True)
                try:
                    length = u.line_length(extended_skeleton)
                    width = u.direct_intersect_distance(extended_skeleton,contour)
                except Exception:
                    width = []
                if np.any(width):
                    l1, l2, profile_mesh, midline = u.straighten_by_orthogonal_lines(contour, extended_skeleton, length, width, unit_micron= 0.5)
                    result, l1, l2 = u.add_poles(l1, l2, pole1, pole2)
                    contour_fit.append(contour)
                    mesh.append(result)             
                    midlines.append(midline)
        
        self.mesh_dataframe = pd.DataFrame({'contour': contour_fit, 'mesh': mesh, 'midline': midlines})
        self.mesh_dataframe['frame'] = self.image_name
        
    def split_cells(self, thresh = 0.3):
        
        self.get_inverted_image()
        print('Splitting cells ...')
        
        new_meshes = []
        new_contours = []
        new_midlines = []
        
        for j in tqdm(range(len(self.mesh_dataframe))):
            
            x1, y1, x2, y2, contour, midline = u.separate_singleframe_meshdata(j, self.mesh_dataframe)
            
            complete_mesh = np.array(np.column_stack((x1, y1, x2, y2)))
            
            step_length_px = u.get_step_length(x1, y1, x2, y2)*self.px
            
            width_px = u.get_width(x1, y1, x2, y2)*self.px
            
            intprofile = u.measure_along_midline_interp2d(midline, self.im_interp2d, width = 5)
            
            weighted_intprofile = u.get_weighted_intprofile(intprofile, width_px)
            
            constrDegree, relPos, constrDegree_abs, ctpos = u.constr_degree_single_cell_min(weighted_intprofile, step_length_px)
            
            if constrDegree > thresh:
                x, y, left, right = u.split_point(x1, y1, x2, y2, ctpos)
                
                mesh1rec = np.concatenate((complete_mesh[:(ctpos-2)], right , np.array([[x, y, x, y]])))
                mesh2rec = np.concatenate((np.array([[x, y, x, y]]), left ,complete_mesh[(ctpos+2):]))
                
                try:
                    mesh1, contour1, midline1 = u.split_mesh2mesh(mesh1rec[:,0], mesh1rec[:,1], mesh1rec[:,2], mesh1rec[:,3])
                    mesh2, contour2, midline2 = u.split_mesh2mesh(mesh2rec[:,0], mesh2rec[:,1], mesh2rec[:,2], mesh2rec[:,3])
                except Exception:
                    continue
                if 4 <= len(mesh1) and len(mesh2) <= 500:
                    new_meshes.append(mesh1)
                    new_meshes.append(mesh2)
                    
                    new_contours.append(contour1)
                    new_contours.append(contour2)
                    
                    new_midlines.append(midline1)
                    new_midlines.append(midline2)
            else:
                new_meshes.append(complete_mesh)
                new_contours.append(contour)
                new_midlines.append(midline)
        self.processed_mesh_dataframe = pd.DataFrame({'mesh': new_meshes, 'contour': new_contours, 'midline': new_midlines})
        self.processed_mesh_dataframe['frame'] = self.image_name

    def object_detection(self, folder_path, channel_suffix = '', log_sigma = 3, kernel_width = 4, min_overlap_ratio=0.01, max_external_ratio=0.1):
                    
        detections = []
        cell_ids = []
        
        for cell in self.cells:
            cell.object_contours = u.get_subcellular_objects(cell.contour, self.channel[channel_suffix], 
                                                             cell.cell_id, log_sigma, 
                                                             kernel_width, min_overlap_ratio, 
                                                             max_external_ratio)
        
            detections.append(cell.object_contours)
            cell_ids.append(cell.cell_id)
        
        self.detections_df = pd.DataFrame({'cell_id' : cell_ids, 'object_contours': detections})
        self.detections_df.insert(0, 'frame', self.frame)
        return self.detections_df
        
        
        
        
class Cell:
    
    def __init__(self, contour, mesh, midline, shape, cell_id):
        
        self.discard = False
        self.contour = contour
        self.mesh = mesh
        self.midline = midline
        self.cell_id = cell_id
        
        self.x1 = self.mesh[:,0]
        self.y1 = self.mesh[:,1]
        self.x2 = self.mesh[:,2]
        self.y2 = self.mesh[:,3]
        
        self.profile_mesh = None
        self.object_contours = None
        
        self.contour_features = {}
        self.phaco_features = {}
        self.nucleoid_features = {}
        self.profiling_data = {}
        self.svm_features = {}    


    


# -*- coding: utf-8 -*-
"""
Created on Wed Apr  5 09:45:56 2023

@author: Bart Steemans. Govers Lab.
"""
from .measurements import *
from . import utilities as u
import numpy as np

class Features:
    
    def __init__(self, image_obj):
        
        self.image_obj = image_obj
    
    def get_all_features(self, cell, add_profiling_data):
        
        self.get_contour_features(cell)
        self.get_phaco_features(cell, add_profiling_data)
        self.get_nucleoid_features(cell, add_profiling_data)
        
    
        
    def get_contour_features(self, cell):
        cell.contour_features = {}
        if 4 <= len(cell.mesh) <= 500:              
            cell.profile_mesh = u.get_profile_mesh(cell.mesh, u.get_width(cell.x1, cell.y1, cell.x2, cell.y2))
            
            self.step_length = get_step_length(cell.x1, cell.y1, cell.x2, cell.y2, self.image_obj.px)
            cell.contour_features['cell_length'] = get_length(self.step_length)
            cell.contour_features['cell_width'], self.width_not_ordered = get_avg_width(cell.x1, cell.y1, cell.x2, cell.y2, self.image_obj.px)
            cell.profiling_data['cell_widthno'] = self.width_not_ordered 
            cell.contour_features['cell_area'] = get_area(cell.contour, self.image_obj.px)
            cell.contour_features['cell_volume'] = get_volume(cell.x1, cell.y1, cell.x2, cell.y2, self.step_length, self.image_obj.px)
            cell.contour_features['cell_surface_area'] = get_surface_area(self.width_not_ordered, self.step_length)
            cell.contour_features['cell_SOV'] = get_surface_area_over_volume(cell.contour_features['cell_surface_area'], cell.contour_features['cell_volume'])
            cell.contour_features['cell_perimeter'], cell.contour_features['cell_circularity'], cell.contour_features['cell_POA'], cell.contour_features['cell_sphericity'] = get_cell_perimeter_measurements(cell.contour, cell.contour_features['cell_area'], self.image_obj.px)
            cell.contour_features['cell_width_variability'] = get_cell_width_variability(self.width_not_ordered)
            cell.contour_features['max_curvature'], cell.contour_features['min_curvature'], cell.contour_features['mean_curvature'] = get_curvature_characteristics(cell.contour, self.image_obj.px)
        else:
            cell.discard = True
            
    def get_phaco_features(self, cell, add_profiling_data):
        cell.phaco_features = {}
        if self.image_obj.im_interp2d is None:
            self.image_obj.get_inverted_image()
        if 4 <= len(cell.mesh) <= 500 and add_profiling_data:    
            try:
                cell.profiling_data['phaco_axial_intensity'] = u.measure_along_midline_interp2d(cell.midline, self.image_obj.im_interp2d, width = 5)
                cell.profiling_data['phaco_midline_intensity'] = u.measure_along_midline_interp2d(cell.midline, self.image_obj.im_interp2d, width = 1)

                cell.profiling_data['phaco_mesh_intensity'] = self.image_obj.im_interp2d.ev(cell.profile_mesh[0],cell.profile_mesh[1]).T
                cell.profiling_data['phaco_average_mesh_intensity'] = u.get_profilemesh_intensity(cell.profiling_data['phaco_mesh_intensity'])
                cell.profiling_data['phaco_weighted_intensity'] = u.get_weighted_intprofile(cell.profiling_data['phaco_axial_intensity'], self.width_not_ordered )
                cell.profiling_data['CD'], cell.profiling_data['relpos'], constrDegree_abs, ctpos = u.constr_degree_single_cell_min(cell.profiling_data['phaco_weighted_intensity'], self.step_length)
                
            except Exception:
                print(f'The mesh of Cell number {cell.cell_id} is incorrect')
                cell.discard = True
        else:
            cell.discard = True
            
    def get_svm_features(self, cell):
        cell.svm_features = {}
        cell.profiling_data = {}
        if self.image_obj.im_interp2d is None:
            self.image_obj.get_inverted_image()
            
        if 4 <= len(cell.mesh) <= 500: 
            try:
                cell.profile_mesh = u.get_profile_mesh(cell.mesh, u.get_width(cell.x1, cell.y1, cell.x2, cell.y2))
                self.step_length = get_step_length(cell.x1, cell.y1, cell.x2, cell.y2, self.image_obj.px)
                
                # Contour is added here for labeling purposes
                cell.svm_features['contour'] = cell.contour
                
                # Morphological features
                cell.svm_features['cell_length'] = get_length(self.step_length)
                cell.svm_features['cell_width'], self.width_not_ordered = get_avg_width(cell.x1, cell.y1, cell.x2, cell.y2, self.image_obj.px)
                cell.svm_features['cell_area'] = get_area(cell.contour, self.image_obj.px)
                cell.svm_features['cell_volume'] = get_volume(cell.x1, cell.y1, cell.x2, cell.y2, self.step_length, self.image_obj.px)
                cell.svm_features['cell_surface_area'] = get_surface_area(self.width_not_ordered, self.step_length)
                cell.svm_features['max_curvature'], cell.svm_features['min_curvature'], cell.svm_features['mean_curvature'] = get_curvature_characteristics(cell.contour, self.image_obj.px)
                cell.svm_features['cell_perimeter'], cell.svm_features['cell_circularity'], cell.svm_features['cell_POA'], cell.svm_features['cell_sphericity'] = get_cell_perimeter_measurements(cell.contour, cell.svm_features['cell_area'], self.image_obj.px)
                
                # Cell surface intensity features
                total_int, max_int, mean_int = get_total_phaco_intensity(cell.contour, self.image_obj.image.shape, self.image_obj.im_interp2d)
                cell.svm_features['phaco_total_intensity'] = total_int
                cell.svm_features['phaco_max_intensity'] = max_int
                cell.svm_features['phaco_mean_intensity'] = mean_int
            
                # Contour intensity features
                cell.profiling_data['phaco_contour_intensity'] = get_contour_intensity(cell.contour, self.image_obj.im_interp2d)
                cell.profiling_data['phaco_contour_intensity_variability'] = measure_contour_variability(cell.profiling_data['phaco_contour_intensity'])
                
                sorted_phaco_contour_intensity = sorted(cell.profiling_data['phaco_contour_intensity'])
                cell.svm_features['phaco_contour_peaks'] = find_signal_peaks(cell.profiling_data['phaco_contour_intensity'], max_int)
                cell.svm_features['phaco_max_contour_intensity'] = np.max(cell.profiling_data['phaco_contour_intensity'])
                cell.svm_features['phaco_mean_contour_intensity'] = np.mean(cell.profiling_data['phaco_contour_intensity'])
                cell.svm_features['phaco_min_contour_intensity'] = np.mean(sorted_phaco_contour_intensity[:10])
                cell.svm_features['phaco_max_contour_variability'] = np.max(cell.profiling_data['phaco_contour_intensity_variability'])
                cell.svm_features['phaco_mean_contour_variability'] = np.mean(cell.profiling_data['phaco_contour_intensity_variability'])
    
                # Midline intensity features
                # If no midline available calculate it with mesh2midline( ).
                cell.profiling_data['phaco_axial_intensity'] = u.measure_along_midline_interp2d(cell.midline, self.image_obj.im_interp2d, width = 5)
                
                cell.svm_features['midline_kurtosis'] = get_kurtosis(cell.profiling_data['phaco_axial_intensity'])
                cell.svm_features['midline_skewness'] = get_skew(cell.profiling_data['phaco_axial_intensity'])
                
                
                # Expanded contour intensity features
                expanded_contour = u.expand_contour(cell.contour, scale = 2)
                eroded_contour = u.erode_contour(cell.contour, scale = 2)
                cell.profiling_data['phaco_expanded_contour_intensity'] = get_contour_intensity(expanded_contour, self.image_obj.im_interp2d)
                cell.profiling_data['phaco_eroded_contour_intensity'] = get_contour_intensity(eroded_contour, self.image_obj.im_interp2d)
                
                cell.svm_features['phaco_max_expanded_contour_intensity'] = np.max(cell.profiling_data['phaco_expanded_contour_intensity'])
                cell.svm_features['phaco_mean_expanded_contour_intensity'] = np.mean(cell.profiling_data['phaco_expanded_contour_intensity'])
                
                #Mesh gradient features
                cell.svm_features['phaco_cell_edge_gradient'] = np.average(cell.profiling_data['phaco_eroded_contour_intensity'] - cell.profiling_data['phaco_expanded_contour_intensity'])
            except Exception:
                print(f'The mesh of Cell number {index} is incorrect')
                cell.discard = True
        else:
            cell.discard = True
            print(f'The mesh of Cell number {cell.cell_id} is too small or large')
            
    def get_nucleoid_features(self, cell, add_profiling_data):
        """
        Under construction.
        Input: channel containing nucleoid information

        Returns
        -------
        Dictionary of features for that cell object.

        """
        if cell.object_contours:
            nucleoid_areas = []  # Initialize a list to store nucleoid areas for each contour
            for contour in cell.object_contours:
                area = get_area(contour, self.image_obj.px)  # Calculate area for each contour
                nucleoid_areas.append(area)
            cell.nucleoid_features['nucleoid_area'] = nucleoid_areas
    
    
    
    



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 12 16:47:19 2023

@author: Bart Steemans. Govers Lab.
"""

import os
import pickle
from . import utilities as u
from tqdm import tqdm
from .image import Image
from .omni import Omnipose
from .curation import Curation
import pandas as pd   
import torch
import multiprocessing


class ImageCollection: 
    
    def __init__(self, image_folder_path = None):
        
        self.images = None
        self.masks = None
        self.channel_images = None
        self.channel_list = None
        self.image_filenames = None
        self.image_folder_path = image_folder_path
        self.name = os.path.basename(self.image_folder_path)
        self.image_objects = []
        
        self.mesh_df_collection = pd.DataFrame()
        
        self.all_features_df = None
        self.svm_features_df = None
        
        
    def load_masks(self):
        self.masks, self.mask_filenames = u.read_tiff_folder(self.image_folder_path + '/masks')
        
    def load_phase_images(self, phase_channel = ''):
        self.images, self.image_filenames, self.paths = u.read_tiff_folder(self.image_folder_path, phase_channel, include_paths = True)
    
    def load_channel_images(self, channel_list):
        self.channel_list = channel_list
        self.channel_images = u.read_channels(self.image_folder_path, self.channel_list)
        
    def create_image_objects(self, channel_list = None, phase_channel=''):
        
        if self.masks is None:
            self.load_masks()
            self.load_phase_images(phase_channel)
            
        if channel_list:
            self.channel_images = self.load_channel_images(channel_list)
            
        self.image_objects = []
    
        for i, image_name in enumerate(self.image_filenames):
            
            image = self.images[i]
            mask = self.masks[i]
            mesh_df = self.get_mesh_dataframe(image_name)
    
            img_obj = self.create_image_object(image, image_name, i, mask, mesh_df)
            
            self.add_channels(img_obj, self.channel_images, i)

            self.image_objects.append(img_obj)
            
    def add_channels(self,img_obj, channel_images, i):
        channel_single_image_dict = {}
        if self.channel_images is not None:  # Check if self.channel_images is not None
            for channel, images in self.channel_images.items():
                channel_single_image_dict[channel] = images[i]
        img_obj.channel = channel_single_image_dict
        
        
    def get_mesh_dataframe(self, image_name):
        
        if not self.mesh_df_collection.empty:
            return self.mesh_df_collection[self.mesh_df_collection['frame'] == image_name].reset_index(drop=True)
        else:
            return None
    
    def create_image_object(self, image, image_name, index, mask, mesh_df):
        img_obj = Image(image, image_name, index, mask, mesh_df)
        if mesh_df is not None:
            img_obj.create_cell_object(verbose=False)
        return img_obj
            
# Methods for batch processing of images ----------------------------------------------  
    def segment_images(self, mask_thresh, minsize, n, model_name = 'bact_phase_omni'):
        omni = Omnipose(self.images, self.paths)
        omni.load_models(model_name)
        
        omni.compiled_process(n, mask_thresh, minsize)
        torch.cuda.empty_cache()
        del omni
    
    def batch_detect_objects(self, folder_path, channel_suffix = '', channel_list = None, log_sigma = 3, kernel_width = 4, 
                             min_overlap_ratio=0.01, max_external_ratio=0.1):
        
        if self.channel_images is None:
            self.load_channel_images(channel_list)
            
            for i, image_obj in enumerate(self.image_objects):
                self.add_channels(image_obj, self.channel_images, i)
                
        print("\nDetecting objects within cell ...")
        
        dfs = (image.object_detection(folder_path, channel_suffix, 
                                      log_sigma, kernel_width, 
                                      min_overlap_ratio, 
                                      max_external_ratio) for image in tqdm(self.image_objects))
        self.object_detection_df = pd.concat(dfs, axis=0)
        
        return self.object_detection_df
    
    def batch_process_mesh(self, pkl_name = None, object_list = None, phase_channel = '', join_thresh = 4, split_thresh = 0.35):
        
        self.mesh_df_collection = pd.DataFrame()
        
        if isinstance(object_list, Image):
            
            print(f'\nProcessing image: {object_list.image_name}')
            
            object_list.join_split_pipeline(join_thresh, split_thresh)
            
            self.mesh_df_collection = getattr(object_list, 'processed_mesh_dataframe')
            
            object_list.create_cell_object(verbose=False)
            
        else:
            
            if object_list is None:
                self.create_image_objects(phase_channel = phase_channel)  
                
                object_list = self.image_objects

            for img_obj in object_list:
                print(f'\nProcessing image: {img_obj.image_name}')
                
                img_obj.join_split_pipeline(join_thresh, split_thresh)
                
                self.mesh_df_collection = pd.concat([self.mesh_df_collection, getattr(img_obj, 'processed_mesh_dataframe')])
                
                img_obj.create_cell_object(verbose=False)

        if pkl_name is not None:
            
            self._to_pickle(pkl_name, self.mesh_df_collection)
            
        self._to_pickle("{}_meshdata.pkl".format(self.name), self.mesh_df_collection)
        
        
    def batch_load_mesh(self, pkl_name, pkl_path = None, phase_channel = ''):
        
        self.mesh_df_collection = pd.DataFrame()
        
        if pkl_path is None:
            
            pkl_path = self.image_folder_path
            
        with open(os.path.join(pkl_path, pkl_name), 'rb') as f:
            self.mesh_df_collection = pickle.load(f)
            
        print(f'\nMeshes are loaded from a pickle file named {pkl_name}')  
        
        self.create_image_objects(phase_channel = phase_channel)
         
        
    def batch_mask2mesh(self, pkl_name = None, object_list = None, phase_channel=''):
        
        self.mesh_df_collection = pd.DataFrame()

        if isinstance(object_list, Image):
            print(f'\nProcessing image: {object_list.image_name}')
            
            object_list.mask2mesh()
            self.mesh_df_collection = getattr(object_list, 'mesh_dataframe')
            object_list.create_cell_object(verbose=False)
        else:
            
            if object_list is None:
                self.create_image_objects(phase_channel=phase_channel)
                object_list = self.image_objects
            
            for img_obj in object_list:
                print(f'\nProcessing image: {img_obj.image_name}')
                
                img_obj.mask2mesh()
                self.mesh_df_collection = pd.concat([self.mesh_df_collection, getattr(img_obj, 'mesh_dataframe')])
                
                img_obj.create_cell_object(verbose=False)
        
        if pkl_name is not None:
            self._to_pickle(pkl_name, self.mesh_df_collection)

        self._to_pickle("{}_meshdata.pkl".format(self.name), self.mesh_df_collection)

    def batch_calculate_features(self, add_profiling_data=True, svm=False):
        
        feature_dfs = []
    
        for image in tqdm(self.image_objects):
            df = image.calculate_features(add_profiling_data, svm)
            feature_dfs.append(df)
    
        combined_df = pd.concat(feature_dfs, axis=0)
    
        if svm:
            self.svm_features_df = combined_df
        else:
            self.all_features_df = combined_df

        return combined_df
    
    def curate_dataset(self, path_to_model, cols = 4):
        
        self.batch_calculate_features(svm=True)
        cur = Curation(self.svm_features_df)
        self.curated_df = cur.compiled_curation(path_to_model, cols)
        
        p1, p0 = cur.get_label_proportions()
        print(f"\nProportion of label 1: {p1}")
        print(f"Proportion of label 0: {p0}")

        for index, row in self.curated_df.iterrows():
            frame = row['frame']
            cell_id = row['cell_id']
            label = row['label']
            
            # Find the Image object in the list based on the 'frame' index
            image = self.image_objects[frame]
            
            # Find the Cell object in the Image object based on the 'cell_id' index
            cell = next((c for c in image.cells if c.cell_id == cell_id), None)
            
            if cell is not None and label == 0:
                # Remove the Cell object from the list of Cells within the Image object
                image.cells.remove(cell)
    
    
    
    def dataframe_to_pkl(self, dataframe = 'features'):
        
        if dataframe == 'features':
            
            self._to_pickle("{}_features.pkl".format(self.name), self.all_features_df)
            
        elif dataframe == 'svm_features':
            
            self._to_pickle("{}_svm_features.pkl".format(self.name), self.svm_features_df)
        
        elif dataframe == 'channel_objects':
            
            self._to_pickle("{}_channel_objects.pkl".format(self.name), self.object_detections)
            

    def _to_pickle(self, pkl_name, data):
        """
        Save data to a pkl file.

        """
        file_path = os.path.join(self.image_folder_path, pkl_name)
        with open(file_path, 'wb') as f:
            pickle.dump(data, f)
    

class Pipeline:
    
    def __init__(self):
        self.exp_folder_path = None
    

    #@profile
    def general_pipeline(self, exp_folder_path, **kwargs):
        
        self.exp_folder_path = exp_folder_path
        
        subfolder_paths = [os.path.join(exp_folder_path, directory) for directory in os.listdir(exp_folder_path) if os.path.isdir(os.path.join(exp_folder_path, directory))]
        print(f'\nProcessed folders are: {subfolder_paths}')
        
        for image_path in subfolder_paths:  
            ic = ImageCollection(image_path)
            
            if 'segment' in kwargs and kwargs['segment']:
                ic.load_phase_images(phase_channel=kwargs.get('phase_channel', ''))
                i = len(ic.images)
                ic.segment_images(kwargs.get('mask_thresh', 1), kwargs.get('minsize', 300), n=range(i))
            
            if 'load_mesh' in kwargs and kwargs['load_mesh']:
                filename = ic.name + '_meshdata.pkl'
                ic.batch_load_mesh(filename, phase_channel=kwargs.get('phase_channel', ''))
            else:
                ic.batch_process_mesh(
                    object_list=None,
                    phase_channel=kwargs.get('phase_channel', ''),
                    join_thresh=kwargs.get('join_thresh', 4),
                    split_thresh=kwargs.get('split_thresh', 0.4)
                )
            
            if 'curation' in kwargs and kwargs['curation'] is not None:
                ic.curate_dataset(kwargs['curation'])
            
            feature_type = kwargs.get('feature_type', 'all')
            if feature_type == 'svm':
                
                print("\nCalculating SVM features ...")
                ic.batch_calculate_features(svm=True)
                ic.dataframe_to_pkl(dataframe='svm_features')
            elif feature_type == 'nucleoid':
                ic.batch_detect_objects(
                    image_path,
                    channel_suffix=kwargs.get('channel_suffix', ''),
                    channel_list = kwargs.get('channel_list'),
                    min_overlap_ratio=kwargs.get('min_overlap_ratio', 0.01),
                    max_external_ratio=kwargs.get('max_external_ratio', 0.1)
                )
                print("\nCalculating nucleoid features ...")
                ic.batch_calculate_features(svm=False)
                ic.dataframe_to_pkl(dataframe='features')
            else:
                print("\nCalculating features")
                ic.batch_calculate_features(svm=False)
                ic.dataframe_to_pkl(dataframe='features')
            
            del ic
         
    def process_image_folder(self, image_path, kwargs):
        ic = ImageCollection(image_path)

        if 'segment' in kwargs and kwargs['segment']:
            print('Segmenting ...')
            ic.load_phase_images(phase_channel=kwargs.get('phase_channel', ''))
            i = len(ic.images)
            ic.segment_images(kwargs.get('mask_thresh', 1), kwargs.get('minsize', 300), n=range(i))

        if 'load_mesh' in kwargs and kwargs['load_mesh']:
            filename = ic.name + '_meshdata.pkl'
            ic.batch_load_mesh(filename, phase_channel=kwargs.get('phase_channel', ''))
        else:
            ic.batch_process_mesh(
                object_list=None,
                phase_channel=kwargs.get('phase_channel', ''),
                join_thresh=kwargs.get('join_thresh', 4),
                split_thresh=kwargs.get('split_thresh', 0.4)
            )

        if 'curation' in kwargs and kwargs['curation'] is not None:
            ic.curate_dataset(kwargs['curation'])

        feature_type = kwargs.get('feature_type', 'all')
        if feature_type == 'svm':
            print("\nCalculating SVM features ...")
            ic.batch_calculate_features(svm=True)
            ic.dataframe_to_pkl(dataframe='svm_features')
        elif feature_type == 'nucleoid':
            ic.batch_detect_objects(
                image_path,
                channel_suffix=kwargs.get('channel_suffix', ''),
                channel_list=kwargs.get('channel_list'),
                min_overlap_ratio=kwargs.get('min_overlap_ratio', 0.01),
                max_external_ratio=kwargs.get('max_external_ratio', 0.1)
            )
            print("\nCalculating nucleoid features ...")
            ic.batch_calculate_features(svm=False)
            ic.dataframe_to_pkl(dataframe='features')
        else:
            print("\nCalculating features")
            ic.batch_calculate_features(svm=False)
            ic.dataframe_to_pkl(dataframe='features')

        del ic

    def general_pipeline_parallel(self, exp_folder_path, **kwargs):
        self.exp_folder_path = exp_folder_path

        subfolder_paths = [os.path.join(exp_folder_path, directory) for directory in os.listdir(exp_folder_path) if os.path.isdir(os.path.join(exp_folder_path, directory))]
        print(f'\nProcessed folders are: {subfolder_paths}')

        # Create a multiprocessing pool with the specified number of cores (default: 2)
        pool = multiprocessing.Pool(kwargs.get('n_cores', 2))

        # Map the function to process each image folder in parallel
        pool.starmap(self.process_image_folder, [(image_path, kwargs) for image_path in subfolder_paths])

        # Close the pool and wait for all processes to complete
        pool.close()
        pool.join()
    
    
    
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Mon May 15 16:08:02 2023

@author: Bart Steemans. Govers Lab.
"""

from cellpose_omni import models, core, io
from cellpose_omni import models
from cellpose_omni.models import MODEL_NAMES

class Omnipose:
    
    def __init__(self, imgs, paths):
        
        self.imgs = imgs
        self.files = paths
        self.use_GPU = core.use_gpu()
        self.masks = None
        self.flows = None
        self.model = None
        print(f'>>> GPU activated? {self.use_GPU}')
        
    def load_models(self, model_name):
        if self.model is None:
            self.model = models.CellposeModel(gpu= self.use_GPU, model_type=model_name)
        
    def process(self , n = [0], mask_thresh = 1, minsize = 300):
        chans = [0,0] #th is means segment based on first channel, no second channel 
        mask_threshold = mask_thresh#negative = mask becomes larger, positive = mask becomes thinner
        verbose = 0 # turn on if you want to see more output 
        use_gpu = self.use_GPU #defined above
        transparency = True # transparency in flow output
        rescale=None # give this a number if you need to upscale or downscale your images
        omni = True # we can turn off Omnipose mask reconstruction, not advised 
        flow_threshold = 0.0 # default is .4, but only needed if there are spurious masks to clean up; slows down output
        resample = True #whether or not to run dynamics on rescaled grid or original grid 
        batch_size = 1
        niter = None
        #diamter: too small -> oversegmentation, too large -> undersegmentation
        #if diameter = None, the diam will be estimated (in images from test-set )
        self.masks, self.flows, styles = self.model.eval([self.imgs[i] for i in n],niter = niter, channels=chans,rescale=rescale,mask_threshold=mask_threshold,transparency=transparency,
                                          flow_threshold=flow_threshold,omni=omni,resample=resample,cluster = True, verbose=verbose, diameter =None, min_size= minsize, batch_size = batch_size)
        
    def save_masks(self):
        io.save_masks(self.imgs, self.masks, self.flows, self.files, 
              png=False,
              tif=True, # whether to use PNG or TIF format
              suffix='', # suffix to add to files if needed 
              save_flows=False, # saves both RGB depiction as *_flows.png and the raw components as *_dP.tif
              save_outlines=False, # save outline images 
              dir_above=0, # save output in the image directory or in the directory above (at the level of the image directory)
              in_folders=True, # save output in folders (recommended)
              save_txt=False, # txt file for outlines in imageJ
              save_ncolor=False) # save ncolor version of masks for visualization and editing 
    
    def compiled_process(self, n, mask_thresh, minsize):
        self.process(n, mask_thresh, minsize)
        self.save_masks()


# -*- coding: utf-8 -*-
"""
Created on Wed Apr  5 16:39:07 2023

@author: Bart Steemans. Govers Lab. 
"""
# Helper functions for extracting features
import numpy as np
import math
from shapely.geometry import Polygon
from skimage.draw import polygon2mask
from scipy.signal import find_peaks
from scipy.stats import kurtosis, skew
# Contour Feature functions ------------------------------------------------------------------------
def get_step_length(x1, y1, x2, y2, px):
    dx = x1[1:] + x2[1:] - x1[:-1] - x2[:-1]
    dy = y1[1:] + y2[1:] - y1[:-1] - y2[:-1]
    return (np.sqrt(dx**2 + dy**2) / 2)*px
def get_length(step_length):
    return np.sum(step_length)

def get_avg_width(x1, y1, x2, y2, px):
    width_not_ordered = np.sqrt((x1-x2)**2+(y1-y2)**2)
    sorted_width = sorted(width_not_ordered, reverse = True)
    width = (sum(sorted_width[:math.floor(len(sorted_width)/3)]) / math.floor(len(sorted_width)/3))
    return width*px, width_not_ordered*px

def get_area(contour, px):
    poly = Polygon(contour)
    area = poly.area
    return area*px*px

def get_volume(x1, y1, x2, y2, step_length, px):
    d = np.sqrt((x1-x2)**2+(y1-y2)**2)
    volume = np.trapz((np.pi*(d/2)**2) , dx = step_length)
    return volume*px*px

def get_surface_area(width_no, step_length):
    widths = width_no[1:]
    surface_areas = 2 * np.pi * (widths / 2) * step_length
    total_surface_area = np.sum(surface_areas)
    return total_surface_area
def get_surface_area_over_volume(sa, vol):
    return (sa/vol)

def get_cell_perimeter_measurements(contour, area, px):
    v1 = contour[:-1]
    v2 = contour[1:]
    d = v2-v1
    perimeter = np.sum(np.sqrt(np.sum(d**2,axis=1)))*px
    circularity = (4*np.pi*area)/(perimeter)**2
    compactness = (perimeter ** 2) / area
    sphericity = (np.pi * 1.5 * (perimeter / (2 * np.pi)) ** 1.5) / area
    return perimeter, circularity, compactness, sphericity

#cell width variability calculated based on the 50% highest cell widths
def get_cell_width_variability(width_no):
    sorted_width = sorted(width_no, reverse = True)
    half_idx = len(sorted_width) // 2
    half_width = sorted_width[:half_idx]
    width_var = np.std(half_width) / np.mean(half_width)    
    return width_var

#ask Sander what the difference is between this and 3 consecutive points (MATLAB code)
def get_curvature_characteristics(contour, px):
    dx = np.gradient(contour[:, 0]*px)
    dy = np.gradient(contour[:, 1]*px)
    d2x = np.gradient(dx)
    d2y = np.gradient(dy)
    curvature = np.abs(dx*d2y - dy*d2x) / (dx**2 + dy**2)**1.5
    max_c = np.max(curvature)
    min_c = np.min(curvature)
    mean_c = np.nanmean(curvature)
    return max_c, min_c, mean_c

def get_total_phaco_intensity(contour, shape, interp2d):
    mask = polygon2mask(shape, contour)
    coords = np.column_stack(np.where(mask)).astype(int)
    values = interp2d.ev(coords[:,0], coords[:,1])
    return np.sum(values), np.max(values), np.mean(values)
    # Calculate the total phase intensity within the contour mask

def get_contour_intensity(contour, interp2d):
    data =  interp2d.ev(contour.T[0], contour.T[1])
    return data

def find_signal_peaks(signal, maximum):
    peaks, _ = find_peaks(signal, prominence= 0.5, height = (maximum * 0.5))
    return len(peaks)

def measure_contour_variability(signal):
    extra_contour_intensities = np.concatenate([signal, signal[:10]])
    contour_intensities_variability = np.array([np.std(extra_contour_intensities[ss:ss+10]) for ss in range(1, len(extra_contour_intensities)-10)])
    return contour_intensities_variability

def get_kurtosis(signal):
    return kurtosis(signal)

def get_skew(signal):
    return skew(signal)


# -*- coding: utf-8 -*-
"""
Created on Thu Apr 13 14:02:27 2023

@author: Bart Steemans. Govers Lab.
"""

import matplotlib.pyplot as plt
import numpy as np

dpi = 300
def plot_contour(cell, image, title = None):
    fig=plt.figure(figsize=(12,12))
    cellobj = image.cells[cell]
    contour = cellobj.contour
    midline = cellobj.midline
    #ori_contour = image.mesh_dataframe['contour'][cell]
    xmin = np.min(contour.T[1]-10)
    xmax = np.max(contour.T[1]+10)
    ymin = np.min(contour.T[0]-10)
    ymax = np.max(contour.T[0]+10)
    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)
    plt.imshow(image.image, cmap = 'gist_gray')
    
    
    plt.plot(contour.T[1], contour.T[0],'-', c = 'w', lw = 5)
    
    
    #plt.plot(ori_contour.T[1], ori_contour.T[0],'-', c = 'y')
    #plt.plot(midline.T[1], midline.T[0], c = 'cyan')
    plt.title(f"Cell {cellobj.cell_id}")
    if title != None:
        plt.title(f'{title}', fontsize = 16)#
    # plt.savefig('C:/Users/u0158103/Documents/PhD/Pictures/contour_plot.png', dpi=dpi)
    plt.show()

def plot_contour_scientific(cell, image,crop_size=300):
    fig=plt.figure(figsize=(12,12))
    cellobj = image.cells[cell]
    contour = cellobj.contour
    midline = cellobj.midline
    
    # Calculate the center of the contour
    center_x = np.mean(contour.T[1])
    center_y = np.mean(contour.T[0])

    # Calculate the new cropping boundaries based on the fixed crop size
    half_crop_size = crop_size / 2
    xmin = center_x - half_crop_size
    xmax = center_x + half_crop_size
    ymin = center_y - half_crop_size
    ymax = center_y + half_crop_size

    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)
    plt.imshow(image.image, cmap='gist_gray')
    
    plt.plot(contour.T[1], contour.T[0],'-', c = 'w', lw = 5)
    
    
    #plt.plot(ori_contour.T[1], ori_contour.T[0],'-', c = 'y')
    #plt.plot(midline.T[1], midline.T[0], c = 'cyan')
    plt.title(f"Cell {cellobj.cell_id}")
    # Add scale bar
    scale_length_um = 1
    scale_length_px = scale_length_um / 0.065  # Convert µm to pixels
    scale_x = (xmin + xmax) / 2 - scale_length_px / 2  # Center the scale bar
    scale_y = ymin + (ymax - ymin) * 0.02  # Adjust the position of the scale bar
    plt.plot([scale_x, scale_x + scale_length_px], [scale_y, scale_y], color='white', lw=5)

    plt.savefig('C:/Users/u0158103/Documents/PhD/Pictures/contour_plot2.svg', dpi=400)
    plt.show()



def plot_objects(cell, image,chann, title = None):
    fig=plt.figure(figsize=(12,12))
    cellobj = image.cells[cell]
    contour = cellobj.contour
    
    xmin = np.min(contour.T[1]-10)
    xmax = np.max(contour.T[1]+10)
    ymin = np.min(contour.T[0]-10)
    ymax = np.max(contour.T[0]+10)
    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)
    plt.imshow(image.channel[chann], cmap = 'gist_gray')
    plt.plot(contour.T[1], contour.T[0],'-', c = 'yellow', lw = 2)
    if cellobj.object_contours is not None:
        for nuc_contour in cellobj.object_contours:
            plt.plot(nuc_contour.T[1], nuc_contour.T[0],'-', c = 'cyan', lw = 2)
            
    plt.title(f"Cell {cellobj.cell_id}")
    plt.axis('off')
    
    plt.show()

def plot_mesh_scientific(cell, image, crop_size = 300):
    fig=plt.figure(figsize=(12,12))
    cellobj = image.cells[cell]
    profile_mesh = cellobj.profile_mesh
    # Calculate the center of the contour
    center_x = np.mean(cellobj.contour.T[1])
    center_y = np.mean(cellobj.contour.T[0])

    # Calculate the new cropping boundaries based on the fixed crop size
    half_crop_size = crop_size / 2
    xmin = center_x - half_crop_size
    xmax = center_x + half_crop_size
    ymin = center_y - half_crop_size
    ymax = center_y + half_crop_size

    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)
    plt.imshow(image.image, cmap='gist_gray')
    plt.plot(cellobj.contour.T[1], cellobj.contour.T[0],'-', c = 'w')
    
    plt.plot([cellobj.y1[::2], cellobj.y2[::2]], [cellobj.x1[::2], cellobj.x2[::2]], c = 'cyan')
    # Add scale bar
    scale_length_um = 1
    scale_length_px = scale_length_um / 0.065  # Convert µm to pixels
    scale_x = (xmin + xmax) / 2 - scale_length_px / 2  # Center the scale bar
    scale_y = ymin + (ymax - ymin) * 0.02  # Adjust the position of the scale bar
    plt.plot([scale_x, scale_x + scale_length_px], [scale_y, scale_y], color='white', lw=5)
    #plt.scatter(profile_mesh[1][::2].T, profile_mesh[0][::2].T, s = 3, c = 'cyan')
    plt.title(f'Cell number: {cellobj.cell_id} with contour and profiling mesh', fontsize = 16)
    plt.savefig('C:/Users/u0158103/Documents/PhD/Pictures/mesh_plot_scientific2.svg', dpi=400)
    plt.show()
    
def plot_mesh(cell, image, title):
    fig=plt.figure(figsize=(12,12))
    cellobj = image.cells[cell]
    profile_mesh = cellobj.profile_mesh
    xmin = np.min(profile_mesh[1]-10)
    xmax = np.max(profile_mesh[1]+10)
    ymin = np.min(profile_mesh[0]-10)
    ymax = np.max(profile_mesh[0]+10)
    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)
    plt.imshow(image.image, cmap = 'gist_gray')
    plt.plot(cellobj.contour.T[1], cellobj.contour.T[0],'-', c = 'w')
    
    plt.plot([cellobj.y1[::2], cellobj.y2[::2]], [cellobj.x1[::2], cellobj.x2[::2]], c = 'cyan')
    #plt.scatter(profile_mesh[1][::2].T, profile_mesh[0][::2].T, s = 3, c = 'cyan')
    plt.title(f'Cell number: {cellobj.cell_id} with contour and profiling mesh', fontsize = 16)
    plt.savefig('C:/Users/u0158103/Documents/PhD/Pictures/mesh_plot.png', dpi=dpi)
    plt.show()
    
def plot_comparison(cell, image):    
    cellobj = image.cells[cell]    
    fig, axs = plt.subplots(1, 2, figsize=(10, 5))
    profile_mesh = cellobj.profile_mesh
    profiling_data = cellobj.profiling_data
    
    
    int_midline = profiling_data['phaco_midline_intensity']
    int_axial = profiling_data['phaco_axial_intensity']
    int_average_mesh = profiling_data['phaco_average_mesh_intensity']
    # weighted_intensity = profiling_data['phaco_weighted_intensity']
    # CD = round(profiling_data['CD'], 3)
    # relpos = round(profiling_data['relpos'], 3)
    xmin = np.min(profile_mesh[1]-10)
    xmax = np.max(profile_mesh[1]+10)
    ymin = np.min(profile_mesh[0]-10)
    ymax = np.max(profile_mesh[0]+10)
    axs[0].set_position([0.05, 0.05, 0.4, 0.9])
    axs[0].imshow(image.image, cmap='gist_gray')
    axs[0].plot(cellobj.contour.T[1], cellobj.contour.T[0],'-', c = 'r')
    axs[0].plot(cellobj.midline.T[1],cellobj.midline.T[0],'-', c = 'y')
    #plt.scatter(profile_mesh[1].T, profile_mesh[0].T, s = 3, c = 'cyan')
    # axs[0].set_title(f'Constriction Degree: {CD} \n Relative Position: {relpos}')
    axs[0].set_xlabel(f'Cell Number: {cell}')
    axs[0].set_xlim(xmin, xmax)
    axs[0].set_ylim(ymin, ymax)
    
    # axs[1].plot(int_axial/25000, c = 'y') 
    # axs[1].plot(weighted_intensity/25000)
    # axs[1].plot(cellobj.contour_features['cell_widthno'], c = 'r')
    
    axs[1].plot(int_midline, c = 'r')
    axs[1].plot(int_axial, c = 'm')
    axs[1].plot(int_average_mesh, c = 'b')
   
    axs[1].set_xlabel('Cell Length')
    axs[1].set_ylabel('Intensity [pixel value]')    
    axs[1].legend(['strip (1) intensity', 'strip (7) intensity', 'profile intensity'], frameon = False)

    # axs[1].set_ylabel('Normalized values')
    # axs[1].legend(['Intensity', 'Weighted Intensity', 'Width'], frameon = False)
    plt.show()

def plot_exp_contour(cell, image):    
    cellobj = image.cells[cell]    
    fig, axs = plt.subplots(1, 2, figsize=(10, 5))
    profile_mesh = cellobj.profile_mesh
    profiling_data = cellobj.profiling_data
    contour_intensity = profiling_data['phaco_expanded_contour_intensity']
    exp_contour = profiling_data['expanded_contour']

    
    # int_traditional = profiling_data['phaco_intensity']
    # int_mesh_midline = profiling_data['phaco_axial_intensity']
    # int_midline = profiling_data['phaco_average_mesh_intensity']
    
    xmin = np.min(profile_mesh[1]-10)
    xmax = np.max(profile_mesh[1]+10)
    ymin = np.min(profile_mesh[0]-10)
    ymax = np.max(profile_mesh[0]+10)
    axs[0].set_position([0.05, 0.05, 0.4, 0.9])
    axs[0].imshow(image.image, cmap='gist_gray')
    axs[0].plot(exp_contour.T[1], exp_contour.T[0],'-', c = 'r')
    axs[0].plot(cellobj.midline.T[1],cellobj.midline.T[0],'-', c = 'y')
    #plt.scatter(profile_mesh[1].T, profile_mesh[0].T, s = 3, c = 'cyan')
    axs[0].set_title('profiling mesh')
    axs[0].set_xlabel(f'Cell Number: {cell}')
    axs[0].set_xlim(xmin, xmax)
    axs[0].set_ylim(ymin, ymax)
    
    #axs[1].plot(int_traditional, c = 'y')
    # axs[1].plot(int_mesh_midline, c = 'k')
    # axs[1].plot(int_midline, c = 'b')
    axs[1].plot(contour_intensity, c = 'b')
    axs[1].set_ylim(0, 27000)

    #axs[1].plot(cellobj.features.width_not_ordered, c = 'cyan')
    #axs[1].legend(['midline intensity', 'strip intensity', 'profile intensity'])
    axs[1].legend(['expaned_contour_intensity'], frameon = False)
    plt.show()
    
def plot_width(cell, image, crop_size = 70):    
    cellobj = image.cells[cell]   
    other_cellobj = image.cells[cell + 44]

    fig1, axs1 = plt.subplots(figsize=(12, 12))
    profile_mesh = cellobj.profile_mesh

    center_x = np.mean(cellobj.contour.T[1])
    center_y = np.mean(cellobj.contour.T[0])

    # Calculate the new cropping boundaries based on the fixed crop size
    half_crop_size = crop_size / 2
    xmin = center_x - half_crop_size
    xmax = center_x + half_crop_size
    ymin = center_y - half_crop_size
    ymax = center_y + half_crop_size

    axs1.set_position([0.1, 0.1, 0.5, 0.8])
    axs1.imshow(image.image, cmap='gist_gray')
    axs1.plot(cellobj.contour.T[1], cellobj.contour.T[0], lw=3, c='orange')

    axs1.set_xlim(xmin, xmax)
    axs1.set_ylim(ymin, ymax)
    # Add scale bar
    scale_length_um = 1
    scale_length_px = scale_length_um / 0.065  # Convert µm to pixels
    scale_x = (xmin + xmax) / 2 - scale_length_px / 2  # Center the scale bar
    scale_y = ymin + (ymax - ymin) * 0.02  # Adjust the position of the scale bar
    axs1.plot([scale_x, scale_x + scale_length_px], [scale_y, scale_y], color='white', lw=5)
    axs1.set_xticks([])
    axs1.set_yticks([])
    axs1.set_title('Cell ' + str(cell), fontname='Arial', fontsize=14)
    
    fig2, axs2 = plt.subplots(figsize=(6, 4))

    # Plot the cell width for the first cell (red curve)
    axs2.plot(cellobj.profiling_data['cell_widthno'], c='orange')
    
    # Plot the cell width for the other cell (green curve)
    axs2.plot(other_cellobj.profiling_data['cell_widthno'], c='b')
    
    # Set x-axis ticks and labels
    newxticks = [np.round(cellobj.contour_features['cell_length'] * (tick / profile_mesh.shape[1]), 1) for tick in axs2.get_xticks()]
    axs2.set_xticks(axs2.get_xticks())
    axs2.set_xticklabels(newxticks, fontname='Arial', fontsize=12)
    
    # Set y-axis ticks and label
    newyticks = [np.round(tick, 1) for tick in axs2.get_yticks()]
    axs2.set_yticks(axs2.get_yticks())
    axs2.set_yticklabels(newyticks, fontname='Arial', fontsize=12)
    
    # Set x-axis to start at 0
    axs2.set_xlim(left=0)
    axs2.set_ylim(bottom=0)
    # Set x and y axis labels
    axs2.set_xlabel('cell length [µm]', fontname='Arial', fontsize=12)
    axs2.set_ylabel('cell width [µm]', fontname='Arial', fontsize=12)

    axs2.legend(['overlapping cell', 'single cell'], frameon=False)

    fig3, axs3 = plt.subplots(figsize=(12, 12))
    other_profile_mesh = other_cellobj.profile_mesh

    center_x = np.mean(other_cellobj.contour.T[1])
    center_y = np.mean(other_cellobj.contour.T[0])

    # Calculate the new cropping boundaries based on the fixed crop size
    half_crop_size = crop_size / 2
    xmin = center_x - half_crop_size
    xmax = center_x + half_crop_size
    ymin = center_y - half_crop_size
    ymax = center_y + half_crop_size

    axs3.set_position([0.1, 0.1, 0.5, 0.8])
    axs3.imshow(image.image, cmap='gist_gray')
    axs3.plot(other_cellobj.contour.T[1], other_cellobj.contour.T[0], lw=3, c='b')

    axs3.set_xlim(xmin, xmax)
    axs3.set_ylim(ymin, ymax)

    axs3.set_xticks([])
    axs3.set_yticks([])
    axs3.set_title('Other Cell', fontname='Arial', fontsize=14)
    scale_length_um = 1
    scale_length_px = scale_length_um / 0.065  # Convert µm to pixels
    scale_x = (xmin + xmax) / 2 - scale_length_px / 2  # Center the scale bar
    scale_y = ymin + (ymax - ymin) * 0.02  # Adjust the position of the scale bar
    axs3.plot([scale_x, scale_x + scale_length_px], [scale_y, scale_y], color='white', lw=5)
    
    fig1.savefig('C:/Users/u0158103/Documents/PhD/Pictures/bad_cell_image.svg', dpi=300)
    fig2.savefig('C:/Users/u0158103/Documents/PhD/Pictures/width.svg', dpi=300)
    fig3.savefig('C:/Users/u0158103/Documents/PhD/Pictures/good_cell_image.svg', dpi=300)
    plt.show()



def plot_signal_profile(cell, image):
    cellobj = image.cells[cell]
    profile_mesh = cellobj.profile_mesh
    profiling_mesh = cellobj.profiling_data['phaco_mesh_intensity']
    fig, ax = plt.subplots(2,1, figsize = (10, 10))
    xmin = np.min(profile_mesh[1]-10)
    xmax = np.max(profile_mesh[1]+10)
    ymin = np.min(profile_mesh[0]-10)
    ymax = np.max(profile_mesh[0]+10)
    ax[0].imshow(profiling_mesh, aspect='auto')
    ax[0].get_xticks()
    xticks = (ax[0].get_xticks())
    newxticks = [np.round(cellobj.contour_features['cell_length'] * (tick / profiling_mesh.shape[1]), 1) for tick in
                 xticks]
    ax[0].set_xticklabels(newxticks, fontname='Arial', fontsize=12)
    ax[0].set_yticks([])
    ax[0].set_ylabel('signal\nstraighten image\n', fontname='Arial', fontsize=12)
    ax[0].set_xlabel('cell length [µm]', fontname='Arial', fontsize=12)
    
    
    ax[1].imshow(image, cmap='gist_gray', aspect = 'auto')
    ax[1].plot(cellobj.contour.T[1], cellobj.contour.T[0],'-', c = 'r')
    ax[1].plot(cellobj.midline.T[1],cellobj.midline.T[0],'-', c = 'y')
    ax[1].set_xlim(xmin, xmax)
    ax[1].set_ylim(ymin, ymax)
    plt.show()


# -*- coding: utf-8 -*-
"""
Created on Wed Feb 22 13:46:15 2023

@author: Bart Steemans. Govers Lab. 
"""
import os
import shutil
import tifffile

import numpy as np
import cv2 as cv
import networkx as nx
import math

import skimage
import scipy.ndimage
from scipy import interpolate
from scipy.ndimage import gaussian_filter1d, gaussian_filter, minimum_filter

from skimage.filters import threshold_otsu, rank
from skimage.draw import polygon2mask
from skimage.morphology import skeletonize, ball
#-------------------------------------------------------------------------------------------------------
#FUNCTIONS
def smallest_path(x_coords, y_coords, start_x, start_y):
    G = nx.Graph()
    for i in range(len(x_coords)):
        G.add_node(i, x=x_coords[i], y=y_coords[i])
    for i in range(len(x_coords)):
        for j in range(i + 1, len(x_coords)):
            G.add_edge(i, j, weight=np.sqrt((x_coords[i] - x_coords[j])**2 + (y_coords[i] - y_coords[j])**2))
    start = None
    for i in range(len(x_coords)):
        if x_coords[i] == start_x and y_coords[i] == start_y:
            start = i
            break
    T = nx.minimum_spanning_tree(G)
    path = nx.dfs_postorder_nodes(T, source=start)
    sorted_points = list(path)
    return [x_coords[i] for i in sorted_points],[y_coords[i] for i in sorted_points]

def skeleton_endpoints(skel):
    # Make our input nice, possibly necessary.
    skel = skel.copy()
    skel[skel!=0] = 1
    skel = np.uint8(skel)
    
    # Apply the convolution.
    kernel = np.uint8([[1,  1, 1],
                       [1, 10, 1],
                       [1,  1, 1]])

    filtered = cv.filter2D(skel, -1 ,kernel)

    out = np.zeros_like(skel)
    out[np.where(filtered==11)] = 1
    return out

def get_ordered_list(points, x, y):
   distance = (((x-points[1])**2 + ( y- points[0])**2)** 0.5)
   args = np.argsort(distance)
   return args

#@jit(nopython=True, cache=True)
def line_intersect(x1, y1, x2, y2, x3, y3, x4, y4):  # Ax+By = C
    A1 = y2 - y1
    B1 = x1 - x2
    C1 = A1 * x1 + B1 * y1
    A2 = y4 - y3
    B2 = x3 - x4
    C2 = A2 * x3 + B2 * y3
    intersect_x = (B1 * C2 - B2 * C1) / (A2 * B1 - A1 * B2)
    intersect_y = (A1 * C2 - A2 * C1) / (B2 * A1 - B1 * A2)
    return intersect_x, intersect_y

def line_contour_intersection(p1, p2, contour):
    v1, v2 = contour[:-1], contour[1:]
    x1, y1 = v1.T
    x2, y2 = v2.T
    x3, y3 = p1
    x4, y4 = p2
    xy = np.array(line_intersect(x1, y1, x2, y2, x3, y3, x4, y4)).T
    dxy_v1 = xy - v1
    dxy_v2 = xy - v2
    dxy = dxy_v1 * dxy_v2
    intersection_points = xy[np.where(np.logical_and(dxy[:, 0] < 0, dxy[:, 1] < 0))]
    if len(intersection_points) > 2:
        dist = np.sum(np.square(np.tile(p1, (len(intersection_points), 1)) - intersection_points),
                      axis=1)
        intersection_points = intersection_points[np.argsort(dist)[0:2]]
    return intersection_points

def unit_perpendicular_vector(data, closed= True):

    p1 = data[1:]
    p2 = data[:-1]
    dxy = p1 - p2
    ang = np.arctan2(dxy.T[1], dxy.T[0]) + 0.5 * np.pi
    dx, dy = np.cos(ang), np.sin(ang)
    unit_dxy = np.array([dx, dy]).T
    if not closed:
        unit_dxy = np.concatenate([[unit_dxy[0]], unit_dxy])
    else:
        unit_dxy = np.concatenate([unit_dxy,[unit_dxy[-1]]])
    return unit_dxy

def get_contour(cell, masks, smoothing = 13):
    temp_mask = np.where( masks == cell, 1, 0)
    contour, coord =  cv.findContours(cv.convertScaleAbs(temp_mask), cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)
    
    contour = np.array(contour)
    
    contourx = contour.squeeze()[:, 1] 
    contoury = contour.squeeze()[:, 0] 
    
    s = int(len(contourx) / smoothing) if len(contourx) > smoothing else 1
    
    tck, u =interpolate.splprep([contourx,contoury], u = None, s = s, per=1)  #3; 12
    u_new = np.linspace(u.min(), u.max(), len(contourx))
    outx,outy = interpolate.splev(u_new, tck, der = 0)
    
    return np.array([outx,outy]).T

def extract_skeleton(contour, masks):
    
    contour = contour.copy()
    contour = contour.astype(int)
    img = polygon2mask(masks.shape, contour)
    # Get the bounding rectangle of the largest contour
    x, y, w, h = cv.boundingRect(contour)
    # Create the cropped image
    padding = 5  # Adjust the padding value as per your needs
    x -= padding
    y -= padding
    w += 2 * padding
    h += 2 * padding
    
    cropped_img = img[x:x+w, y:y+h]
    
    # Apply skeletonization on the cropped image
    skeleton = skeletonize(cropped_img, method='lee')
    
    # Retrieve the skeleton coordinates in the original image space
    skeleton_coords = np.column_stack(np.where(skeleton)).astype(int)
    skeleton_coords[:, 0] += x  # Add the x-offset
    skeleton_coords[:, 1] += y  # Add the y-offset
    
    x_skel = skeleton_coords[:, 0]
    y_skel = skeleton_coords[:, 1]
    
    skel_mask = np.zeros_like(masks)
    skel_mask[x_skel,y_skel] = 1
    
    #Sort coordinates based on distance
    skel_end = skeleton_endpoints(skel_mask)
    end_coords = np.argwhere(skel_end == 1)[0]
    sorted_skelx,sorted_skely = (smallest_path(x_skel, y_skel, end_coords[0], end_coords[1]))
    #Smoothen the skeleton
    if len(y_skel)<=3:
        return []
    elif 4 <=len(y_skel)<= 13:
        fkc, u = interpolate.splprep([sorted_skelx,sorted_skely], k=1, s=1, per=0)
    else:
        fkc, u = interpolate.splprep([sorted_skelx,sorted_skely], k=2, s=10, per=0)
    smoothed_skel = np.asarray(interpolate.splev(np.linspace(u.min(), u.max(), 200), fkc, der=0)).T
    return smoothed_skel

def find_poles(smoothed_skeleton,
               smoothed_contour,
               find_pole1=True,
               find_pole2=True):
    # find endpoints and their nearest neighbors on a midline
    length = len(smoothed_skeleton)
    extended_pole1 = [smoothed_skeleton[0]]
    extended_pole2 = [smoothed_skeleton[-1]]
    i = 0
    j = 0
    if find_pole1:
        for i in range(10):
            p1 = smoothed_skeleton[i]
            p2 = smoothed_skeleton[i + 1]
            # find the two intersection points between
            # the vectorized contour and line through pole1
            intersection_points_pole1 = line_contour_intersection(p1, p2, smoothed_contour)
            # find the interesection point with the same direction as the outward pole vector
            dxy_1 = p1 - p2
            p1_tile = np.tile(p1, (len(intersection_points_pole1), 1))
            p1dot = (intersection_points_pole1 - p1_tile) * dxy_1
            index_1 = np.where((p1dot[:, 0] > 0) & (p1dot[:, 1] > 0))[0]
            if len(index_1) > 0:
                extended_pole1 = intersection_points_pole1[index_1]
                break
    else:
        i = 1

    if find_pole2:
        for j in range(10):
            p3 = smoothed_skeleton[-1 - j]
            p4 = smoothed_skeleton[-2 - j]
            # find the two intersection points between
            # the vectorized contour and line through pole1
            intersection_points_pole2 = line_contour_intersection(p3, p4, smoothed_contour)
            # find the interesection point with the same direction as the outward pole vector
            dxy_2 = p3 - p4
            p3_tile = np.tile(p3, (len(intersection_points_pole2), 1))
            p3dot = (intersection_points_pole2 - p3_tile) * dxy_2
            index_2 = np.where((p3dot[:, 0] > 0) & (p3dot[:, 1] > 0))[0]
            if len(index_2) > 0:
                extended_pole2 = intersection_points_pole2[index_2]
                break
    else:
        j = 1
    trimmed_midline = smoothed_skeleton[i:length - j]
    return extended_pole1, extended_pole2, trimmed_midline

def extend_skeleton(smoothed_skeleton, smoothed_contour,
                    find_pole1=True, find_pole2=True):
    # initiate approximated tip points
    new_pole1, new_pole2, trimmed_skeleton = find_poles(smoothed_skeleton,
                                                         smoothed_contour,
                                                         find_pole1=find_pole1,
                                                         find_pole2=find_pole2)
    extended_skeleton = np.concatenate([new_pole1,
                                        trimmed_skeleton,
                                        new_pole2])

    extended_skeleton = spline_approximation(extended_skeleton, n = 200, smooth_factor = 1, closed = False)
    return extended_skeleton, new_pole1, new_pole2                           

def distance_matrix(data1, data2):
    x1, y1 = data1.T
    x2, y2 = data2.T
    dx = x1[:, np.newaxis] - x2
    dy = y1[:, np.newaxis] - y2
    dxy = np.sqrt(dx ** 2 + dy ** 2)
    return dxy

def intersect_matrix(line, contour,
                     orthogonal_vectors=None):
    if orthogonal_vectors is None:
        dxy = unit_perpendicular_vector(line, closed=False)
    else:
        dxy = orthogonal_vectors
    v1, v2 = contour[:-1], contour[1:]
    x1, y1 = v1.T
    x2, y2 = v2.T
    x3, y3 = line.T
    perp_xy = line + dxy
    x4, y4 = perp_xy.T
    A1 = y2 - y1
    B1 = x1 - x2
    C1 = A1 * x1 + B1 * y1
    A2 = y4 - y3
    B2 = x3 - x4
    C2 = A2 * x3 + B2 * y3

    A1B2 = A1[:, np.newaxis] * B2
    A1C2 = A1[:, np.newaxis] * C2
    B1A2 = B1[:, np.newaxis] * A2
    B1C2 = B1[:, np.newaxis] * C2
    C1A2 = C1[:, np.newaxis] * A2
    C1B2 = C1[:, np.newaxis] * B2
    #can give outliers when B1A2 = A1B2, division by zero or close to zero
    intersect_x = (B1C2 - C1B2) / (B1A2 - A1B2)
    intersect_y = (A1C2 - C1A2) / (A1B2 - B1A2)
    return intersect_x, intersect_y

def spline_approximation(init_contour, n=200, smooth_factor=1, closed = True):
    if closed:
        tck, u = interpolate.splprep(init_contour.T, u=None, s=smooth_factor, per=1)
    else:
        tck, u = interpolate.splprep(init_contour.T, u=None, s=smooth_factor)
    u_new = np.linspace(u.min(), u.max(), n)
    x_new, y_new = interpolate.splev(u_new, tck, der=0)
    return np.array([x_new, y_new]).T
#min dist has influence on extreme points (lower works good for some cases)
def orthogonal_intersection_point(midline,
                                  outerline,
                                  precomputed_orthogonal_vector=None,
                                  min_dist=1e-100):
    v1, v2 = outerline[:-1], outerline[1:]
    skel_x, skel_y = midline.T
    if precomputed_orthogonal_vector is None:
        intersect_x, intersect_y = intersect_matrix(midline, outerline)
    else:
        intersect_x, intersect_y = intersect_matrix(midline, outerline,
                                                    orthogonal_vectors=precomputed_orthogonal_vector)
    dx_v1 = intersect_x - v1.T[0][:, np.newaxis]
    dx_v2 = intersect_x - v2.T[0][:, np.newaxis]
    dy_v1 = intersect_y - v1.T[1][:, np.newaxis]
    dy_v2 = intersect_y - v2.T[1][:, np.newaxis]
    dx = dx_v1 * dx_v2
    dy = dy_v1 * dy_v2

    dist_x = skel_x[np.newaxis, :] - intersect_x
    dist_y = skel_y[np.newaxis, :] - intersect_y
    #influence on extreme points
    non_bounadry_points = np.where(np.logical_and(dy >= 0, dx >= 0))
    dist_matrix = np.sqrt(dist_x ** 2 + dist_y ** 2)
    dist_matrix[non_bounadry_points] = np.inf
    dist_matrix[dist_matrix <= min_dist] = np.inf
    
    nearest_id_x = np.argsort(dist_matrix, axis=0)[:1]
    nearest_id_y = np.linspace(0, dist_matrix.shape[1]-1, dist_matrix.shape[1]).astype(int)
    pos_list = np.array([intersect_x[nearest_id_x[0], nearest_id_y],
                         intersect_y[nearest_id_x[0], nearest_id_y]]).T
    return pos_list

def straighten_by_orthogonal_lines(contour, midline, length, width, unit_micron= 0.5):
    # estimate profile mesh size
    median_width = np.median(width)
    N_length = int(round(length / unit_micron))
    N_width = int(round(median_width / unit_micron))
    midline = spline_approximation(midline, n= N_length, smooth_factor=1, closed=False)

    # divide contour
    # if contour[-1] == contour[-2]:
    #     contour = contour[:-1]
    half_contour_1, half_contour_2 = divide_contour_by_midline(midline, contour)#[:-1]
    # infer orthogonal vectors
    ortho_unit_vectors = unit_perpendicular_vector(midline)
    # generate orthogonal profile lines for all midline points except for the polar ones
    l1 = orthogonal_intersection_point(midline, half_contour_1,
                                       precomputed_orthogonal_vector= ortho_unit_vectors)

    l2 = orthogonal_intersection_point(midline, half_contour_2,
                                       precomputed_orthogonal_vector=ortho_unit_vectors)

    dl = (l2 - l1) / N_width
    mult_mat = np.tile(np.arange(N_width + 1), (len(l1), 1))
    mat_x = l1[:, 0][:, np.newaxis] + mult_mat * dl[:, 0][:, np.newaxis]
    mat_y = l1[:, 1][:, np.newaxis] + mult_mat * dl[:, 1][:, np.newaxis]
    profile_mesh = np.array([mat_x, mat_y])
    return l1, l2, profile_mesh, midline



def divide_contour_by_midline(midline, contour):
    dist1 = distance_matrix(contour, midline[0]).flatten()
    dist2 = distance_matrix(contour, midline[-1]).flatten()

    id1, id2 = np.argsort(dist1)[:2]
    id3, id4 = np.argsort(dist2)[:2]

    contour_cp = contour.copy()
    if max(id1, id2) < max(id3, id4):
        term_p1 = max(id1, id2)
        if abs(id3 - id4) == 1:
            term_p2 = max(id3, id4) + 1
        elif abs(id3 - id4) > 1:
            term_p2 = max(id3, id4) + 2
        contour_cp = np.insert(contour_cp, term_p1, midline[0], axis=0)
        contour_cp = np.insert(contour_cp, term_p2, midline[-1], axis=0)

    else:
        term_p1 = max(id3, id4)
        if abs(id1 - id2) == 1:
            term_p2 = max(id1, id2) + 1
        elif abs(id1 - id2) > 1:
            term_p2 = max(id1, id2) + 2
        contour_cp = np.insert(contour_cp, term_p1, midline[-1], axis=0)
        contour_cp = np.insert(contour_cp, term_p2, midline[0], axis=0)

    if term_p1 == term_p2:
        raise ValueError('Two endpoints are identical!')
    else:
        pos1, pos2 = sorted([term_p1, term_p2])
        #print(id1, id2, id3, id4, pos1, pos2, len(contour_cp))
        half_contour_1 = contour_cp[pos1:min(pos2 + 1, len(contour_cp) - 1)]
        half_contour_2 = np.concatenate([contour_cp[pos2:], contour_cp[:pos1 + 1]])
    return half_contour_1, half_contour_2

def line_length(line):
    v1 = line[:-1]
    v2 = line[1:]
    d = v2-v1
    return np.sum(np.sqrt(np.sum(d**2,axis=1)))

def direct_intersect_distance(skeleton, contour):
    v1, v2 = contour[:-1], contour[1:]
    skel_x, skel_y = skeleton[1:-1].T
    intersect_x, intersect_y = intersect_matrix(skeleton[1:-1], contour)
    dx_v1 = intersect_x - v1.T[0][:, np.newaxis]
    dx_v2 = intersect_x - v2.T[0][:, np.newaxis]
    dy_v1 = intersect_y - v1.T[1][:, np.newaxis]
    dy_v2 = intersect_y - v2.T[1][:, np.newaxis]
    dx = dx_v1 * dx_v2
    dy = dy_v1 * dy_v2
    dist_x = skel_x[np.newaxis, :] - intersect_x
    dist_y = skel_y[np.newaxis, :] - intersect_y

    non_boundry_points = np.where(np.logical_and(dy > 0, dx > 0))
    dist_matrix = np.sqrt(dist_x ** 2 + dist_y ** 2)
    dist_matrix[non_boundry_points] = np.inf
    nearest_id_x = np.argsort(dist_matrix, axis=0)[:2]
    nearest_id_y = np.linspace(0, dist_matrix.shape[1] - 1, dist_matrix.shape[1]).astype(int)
    dists = dist_matrix[nearest_id_x[0], nearest_id_y] + dist_matrix[nearest_id_x[1], nearest_id_y]
    return np.concatenate([[0], dists, [0]])

def distance(v1, v2):
    #Euclidean distance of two points
    return np.sqrt(np.sum((np.array(v1) - np.array(v2)) ** 2))

def measure_width(extended_skeleton, smoothed_contour):
    length = line_length(extended_skeleton)
    d_perp = unit_perpendicular_vector(extended_skeleton, closed=False)
    width_list = []
    for i in range(1, len(extended_skeleton)-1):
        xy = line_contour_intersection(extended_skeleton[i],
                                       d_perp[i] + extended_skeleton[i],
                                       smoothed_contour)
        coords = np.average(xy, axis=0)
        if (len(xy) == 2) and (np.isnan(coords).sum() == 0):
            width_list.append(distance(xy[0], xy[1]))
        else:
            raise ValueError('Error encountered while computing line intersection points!')
    return np.array([0]+width_list+[0]), length

def add_poles(coords1, coords2, pole1, pole2):
    coords1 = np.vstack([pole1, coords1[1:-1], pole2])
    coords2 = np.vstack([pole1, coords2[1:-1], pole2])
    result = np.stack([coords1[:,0], coords1[:,1], coords2[:,0], coords2[:,1]], axis=1)
    return result, coords1, coords2

def get_cell_pairs(pole_1, pole_2, cell_ID, maxdist = None):
    dm11 = distance_matrix(pole_1, pole_1)
    dm22 = distance_matrix(pole_2, pole_2)
    dm12 = distance_matrix(pole_1, pole_2)
    for i in cell_ID:
        dm11 = np.insert(dm11, i-1, values=np.zeros((1, dm11.shape[1]), dtype=int), axis=0)
        dm11 = np.insert(dm11, i-1, values=np.zeros(dm11.shape[0], dtype=int), axis=1)
        dm22 = np.insert(dm22, i-1, values=np.zeros((1, dm22.shape[1]), dtype=int), axis=0)
        dm22 = np.insert(dm22, i-1, values=np.zeros(dm22.shape[0], dtype=int), axis=1)
        dm12 = np.insert(dm12, i-1, values=np.zeros((1, dm12.shape[1]), dtype=int), axis=0)
        dm12 = np.insert(dm12, i-1, values=np.zeros(dm12.shape[0], dtype=int), axis=1)
    if maxdist != None:
        maxdistance = maxdist
        neighboring_cells_dm11 = np.argwhere(np.logical_and(dm11 > 0, dm11 <= maxdistance))
        neighboring_cells_dm22 = np.argwhere(np.logical_and(dm22 > 0, dm22 <= maxdistance))
        neighboring_cells_dm12 = np.argwhere(np.logical_and(dm12 > 0, dm12 <= maxdistance))
        nc_concat = np.concatenate((neighboring_cells_dm11, neighboring_cells_dm22, neighboring_cells_dm12), axis=0)
        unique_nc1 = np.unique(nc_concat, axis=0)
        unique_nc2 = unique_nc1[unique_nc1[:, 0] != unique_nc1[:, 1]]
    else:
        #get index list of cell couples. Loop decides the distance for when more than 2 cells from couples with other cells
        maxdistance = 1
        while True: #or maxdist = 8
            neighboring_cells_dm11 = np.argwhere(np.logical_and(dm11 > 0, dm11 <= maxdistance))
            neighboring_cells_dm22 = np.argwhere(np.logical_and(dm22 > 0, dm22 <= maxdistance))
            neighboring_cells_dm12 = np.argwhere(np.logical_and(dm12 > 0, dm12 <= maxdistance))
            nc_concat = np.concatenate((neighboring_cells_dm11, neighboring_cells_dm22, neighboring_cells_dm12), axis=0)
            unique_nc1 = np.unique(nc_concat, axis=0)
            unique_nc2 = unique_nc1[unique_nc1[:, 0] != unique_nc1[:, 1]]
            unique_values, inverse_indices = np.unique(unique_nc2, return_inverse=True)
            counts = np.bincount(inverse_indices)
            # Check if the current neighboring cells are already in the set of unique cells
            if counts == []:
                print("No neighboring cells present in the image")
                break
            try:
                if np.max(counts) > 2:
                    print(f"Found triple neighboring cells after {maxdistance} distance units")
                    maxdistance = maxdistance
                    break
            except ValueError:
                pass
                    
            
            # Increment the maximum distance to search for
            maxdistance += 1
            if maxdistance > 10:
                print('pole distance to reach 4 stitched cells exceeds 20')
                break
            if counts == []:
                unique_nc2 = []
            else:
                neighboring_cells_dm11 = np.argwhere(np.logical_and(dm11 > 0, dm11 <= maxdistance))
                neighboring_cells_dm22 = np.argwhere(np.logical_and(dm22 > 0, dm22 <= maxdistance))
                neighboring_cells_dm12 = np.argwhere(np.logical_and(dm12 > 0, dm12 <= maxdistance))
                nc_concat = np.concatenate((neighboring_cells_dm11, neighboring_cells_dm22, neighboring_cells_dm12), axis=0)
                unique_nc1 = np.unique(nc_concat, axis=0)
                unique_nc2 = unique_nc1[unique_nc1[:, 0] != unique_nc1[:, 1]]
    return unique_nc2

def mesh2contour(x1, y1, x2, y2):
    x2f = np.flip(x2)
    y2f = np.flip(y2)
    # Concatenate the x and y coordinates
    xspline = np.concatenate((x2f[1:], x1[1:]))
    yspline = np.concatenate((y2f[1:], y1[1:]))

    tck, u = interpolate.splprep(np.array([xspline, yspline]), k=3, s=2, per = 1)
    u_new = np.linspace(u.min(), u.max(), 200)
    outx, outy = interpolate.splev(u_new, tck)
    
    return np.array([outx, outy]).T

def split_point(x1, y1, x2, y2, ctpos):
    dx = x2 - x1
    dy = y2 - y1
    dist = np.sqrt(dx**2 + dy**2)
    # ctpos is the index at which the cell is most constricted and where the algorith will split the cell
    xc1 = x1[ctpos]
    yc1 = y1[ctpos]
    xc2 = x2[ctpos]
    yc2 = y2[ctpos]
    newpole_x = (xc1 + xc2) / 2
    newpole_y = (yc1 + yc2) / 2
    
    xt1, yt1 = x1[ctpos+1], y1[ctpos+1]
    distt = dist[ctpos+1]
    # Use indexing to get the x and y coordinates separately
    xt13, yt13 = np.array([xt1, yt1]) + (distt/3) * np.array([dx[ctpos+1], dy[ctpos+1]]) / distt
    xt23, yt23 = np.array([xt1, yt1]) + (distt*(2/3)) * np.array([dx[ctpos+1], dy[ctpos+1]]) / distt
    
    xs1, ys1 = x1[ctpos-1], y1[ctpos-1]
    dists = dist[ctpos-1]
    
    # Use indexing to get the x and y coordinates separately
    xs13, ys13 = np.array([xs1, ys1]) + (dists/3) * np.array([dx[ctpos-1], dy[ctpos-1]]) / dists
    xs23, ys23 = np.array([xs1, ys1]) + (dists*(2/3)) * np.array([dx[ctpos-1], dy[ctpos-1]]) / dists
    
    return newpole_x, newpole_y, np.array([[xt13, yt13, xt23, yt23]]), np.array([[xs13, ys13, xs23, ys23]])

# calculates the constriction degree in relative and absolute value, 
# and the relative position along the length of the cell
# absolute constr degree needs to be multiplied by the width
def constr_degree_single_cell_min(intensity, new_step_length):
    # if intensity == []:
    #     minsize = 0
    #     minsizeabs = 0
    #     ctpos = []
    # else:
    minima = np.concatenate(([False], (intensity.T[1:-1] < intensity.T[:-2]) & (intensity.T[1:-1] < intensity.T[2:]), [False]))
    if all(not x for x in minima) or sum(intensity.T) == 0: 
        minsize = 0
        minsizeabs = 0
        ctpos = []
    else:
        index = np.where(minima)[0]
        dh = np.zeros(index.shape)
        dhi = np.zeros(index.shape)
        hgt = np.zeros(index.shape)
        for i in range(len(index)):
            k = index[i]
            half1 = intensity.T[:k-1]
            half2 = intensity.T[k+1:]
            try:
                dh1 = np.max(half1)-intensity.T[k]
                dh2 = np.max(half2)-intensity.T[k]
                dh[i] = np.min([dh1, dh2])
                dhi[i] = np.mean([dh1, dh2])
                hgt[i] = intensity.T[k]+dhi[i]
            except ValueError:
                minsize = 0
                minsizeabs = 0
                ctposr = []
            fix = np.argmax(dh)
            minsizeabs = dhi[fix]
            minsize = minsizeabs / hgt[fix]
            ctpos = index[fix]
            ctposr = np.cumsum(new_step_length)[ctpos] / np.sum(new_step_length)
        if not minsize:
            minsize = 0
            minsizeabs = 0
    constrDegree = minsize
    constrDegree_abs = minsizeabs
    if not ctpos:
        ctposr = np.nan
    relPos = ctposr
    return constrDegree, relPos, constrDegree_abs, ctpos

# This function calculates the interpolated intensity values at the position of each midline
# This is required for creating the intensity-profile
def intensity_profile(midline, width, ff):
    # x = (x1 + x2) / 2
    # y = (y1 + y2) / 2
    # create the interpolation function using bilinear interpolation
    # interpolate the image at the points given by the mesh grid
    # prf = ff.ev(x, y)
    prf = ff.ev(midline.T[0], midline.T[1])
    sigma = 2  # standard deviation of Gaussian filter
    prf = gaussian_filter1d(prf, sigma)  # apply Gaussian filter
    weighted_prf = prf * width
    return prf, weighted_prf

def get_profilemesh_intensity(profile_mesh):
    prf = np.average(profile_mesh, axis = 0)
    sigma = 2  # standard deviation of Gaussian filter
    prf = gaussian_filter1d(prf, sigma)  # apply Gaussian filter
    return prf

# This function iterates over each segment and sums up the pixel intensities within that segment
# this function is not used in calculating the intensity profile
def measure_intensity(x1, y1, x2, y2, step_length, image, bgr, width):
    #weight the intensity with the length of each branch
    mesh = np.array([x1, y1, x2, y2], dtype=np.int32)
    step_intensity = []
    bgr_image = np.maximum(image - bgr, 0)
    for i in range(1, len(x1)):
        rect = np.array([mesh[:, i - 1], mesh[:, i]])
        rect_reshaped = rect.reshape(-1, 2)
        rect_reshaped = rect_reshaped[np.lexsort((rect_reshaped[:, 1], rect_reshaped[:, 0]))]
        rect_reshaped = rect_reshaped.reshape(-1, 2)
        rect_reshaped = np.append(rect_reshaped, rect_reshaped[0]).reshape(-1, 2)
        x_coords = rect_reshaped[:,1]
        y_coords = rect_reshaped[:,0]
        
        rr, cc = skimage.draw.polygon(y_coords, x_coords)
        mask = np.zeros_like(bgr_image, dtype= bool)
        mask[rr, cc] = True
        intensity = np.sum(bgr_image[mask])
        step_intensity.append(intensity)
    sigma = 2  # standard deviation of Gaussian filter
    inten = gaussian_filter1d(step_intensity, sigma)  # apply Gaussian filter
    weighted_inten = inten * width[:-1]
    return np.array(weighted_inten).T, np.array(inten).T

def interp2d(image):
    ff = interpolate.RectBivariateSpline(range(image.shape[0]), range(image.shape[1]), image, kx=1, ky=1)
    return ff

# def phase_background(image, se_size=3, bgr_erode_num=1):
#     cropped_image = image[int(image.shape[0]*0.05):int(image.shape[0]*0.95), int(image.shape[1]*0.05):int(image.shape[1]*0.95)]
#     thres = skimage.filters.threshold_otsu(cropped_image)
#     mask = cv.threshold(image, thres, 255, cv.THRESH_BINARY_INV)[1]
#     kernel = cv.getStructuringElement(cv.MORPH_RECT, (se_size, se_size))
#     mask = cv.erode(mask, kernel, iterations=bgr_erode_num)
#     bgr_mask = mask.astype(bool)
#     bgr = np.mean(image[bgr_mask])
#     return bgr

def phase_background(image, cell_mask):
    background_mask = np.logical_not(cell_mask)
    bgr = np.mean(image[background_mask])
    return bgr

def get_bbox(x1, y1, x2, y2):
    mesh1 = [x1, x2]
    mesh2 = [y1, y2]
    ymin = int(np.min(mesh1)-10)
    ymax = int(np.max(mesh1)+10)
    xmin = int(np.min(mesh2)-10)
    xmax = int(np.max(mesh2)+10)
    # Crop the image using the boundaries
    return (xmin, xmax, ymin, ymax)

def get_step_length(x1, y1, x2, y2):
    dx = x1[1:] + x2[1:] - x1[:-1] - x2[:-1]
    dy = y1[1:] + y2[1:] - y1[:-1] - y2[:-1]
    return (np.sqrt(dx**2 + dy**2) / 2)

def get_image_for_frame(df_nr, images):
    return images[df_nr]

def separate_meshdata(df_nr, cell_nr, data):
    x1 = data[df_nr]['mesh'][cell_nr][:,0]
    y1 = data[df_nr]['mesh'][cell_nr][:,1]
    x2 = data[df_nr]['mesh'][cell_nr][:,2]
    y2 = data[df_nr]['mesh'][cell_nr][:,3]
    contour = data[df_nr]['contour'][cell_nr]
    return x1, y1, x2, y2, contour

def separate_singleframe_meshdata(cell_nr, data):
    x1 = data['mesh'][cell_nr][:,0]
    y1 = data['mesh'][cell_nr][:,1]
    x2 = data['mesh'][cell_nr][:,2]
    y2 = data['mesh'][cell_nr][:,3]
    contour = data['contour'][cell_nr]
    midline = data['midline'][cell_nr]
    return x1, y1, x2, y2, contour, midline

def distribute_tiff_files(folder_path, batch_size=5):
    tiff_files = [f for f in os.listdir(folder_path) if f.lower().endswith(('.tif', '.tiff'))]
    
    if len(tiff_files) <= batch_size:
        print("No need to distribute files, as there are 5 or fewer TIFF files.")
        return

    # Calculate the number of new folders needed
    num_folders = len(tiff_files) // batch_size
    if len(tiff_files) % batch_size != 0:
        num_folders += 1

    # Create new folders with incremental names
    for i in range(num_folders):
        new_folder_name = f"{folder_path}_{i + 1}"
        os.makedirs(new_folder_name, exist_ok=True)

        # Copy batch_size TIFF files to the new folder or fewer if it's the last folder
        start_index = i * batch_size
        end_index = min((i + 1) * batch_size, len(tiff_files))
        
        for j in range(start_index, end_index):
            source_file = os.path.join(folder_path, tiff_files[j])
            destination_file = os.path.join(new_folder_name, tiff_files[j])
            shutil.copy2(source_file, destination_file)

    print(f"{len(tiff_files)} TIFF files distributed into {num_folders} folders with a batch size of {batch_size}.")


def read_channels(folder_path, channel_list=None):
    if channel_list is None:
        # If no suffixes are provided, assume an empty list
        channel_list = []

    # Get a list of all TIFF files in the folder matching the provided suffixes
    images_dict = {}  # Create a dictionary to store channel images with specified suffixes as keys
    
    for channel in channel_list:
        matching_files = [f for f in os.listdir(folder_path) if f.endswith(f'{channel}.tif') or f.endswith(f'{channel}.tiff')]
        
        if matching_files:
            images_dict[channel] = [tifffile.imread(os.path.join(folder_path, tiff_file)) for tiff_file in matching_files]

    if not any(images_dict.values()):
        raise ValueError(f'\nNo images with specified suffixes ({channel_list}) found in folder: {folder_path}')

    print(f'\nImages with specified suffixes ({channel_list}) found in folder: {folder_path}')
    
    return images_dict


def read_tiff_folder(folder_path, suffix='', include_paths=False):
    # Get a list of all TIFF files in the folder
    tiff_files = [f for f in os.listdir(folder_path) if f.endswith(f'{suffix}.tif') or f.endswith(f'{suffix}.tiff')]
    
    if not tiff_files:
        # Raise an exception if no TIFF files are found in the folder
        raise ValueError('\nNo TIFF files found in folder: ' + folder_path)
    
    images = []
    file_names = []
    paths = [] if include_paths else None

    for tiff_file in tiff_files:
        tiff_path = os.path.join(folder_path, tiff_file)
        image = tifffile.imread(tiff_path)
        images.append(image)
        file_names.append(tiff_file)
        if include_paths:
            paths.append(tiff_path)

    if len(images) == 1:
        print(f'\nOnly one TIFF file found in folder: {folder_path}')
        images = np.expand_dims(images[0], axis=0)
        file_names = [file_names[0]]
        if include_paths:
            paths = [paths[0]]
        
    else:
        print(f'\n{len(images)} TIFF files found in folder: {folder_path}')

    if include_paths:
        return np.array(images), file_names, paths
    else:
        return np.array(images), file_names
        
# Better use the midline attribute instead of the mesh for such purposes
def mesh2midline(x1, y1, x2, y2):
    x = (x1 + x2) / 2
    y = (y1 + y2) / 2
    line = np.array([x, y]).T
    line = spline_approximation(line, n = len(x1), smooth_factor = 3, closed = False)
    return line

def get_weighted_intprofile(intensity, width):
    if intensity.shape[0] == width.shape[0]:
        return intensity * width
    else:
        return []
    
def measure_along_midline_interp2d(midline, im_interp2d, width = 7, subpixel = 0.5):

    unit_dxy = unit_perpendicular_vector(midline, closed=False)
    width_normalized_dxy = unit_dxy * subpixel
    
    data = im_interp2d.ev(midline.T[0], midline.T[1])
    for i in range(1, 1+int(width * 0.5 / subpixel)):
        dxy = width_normalized_dxy * i
        v1 = midline + dxy
        v2 = midline - dxy
        p1 = im_interp2d.ev(v1.T[0], v1.T[1])
        p2 = im_interp2d.ev(v2.T[0], v2.T[1])
        data = np.vstack([p1, data, p2])
    prf = np.average(data, axis=0)
    sigma = 2  # standard deviation of Gaussian filter
    prf = gaussian_filter1d(prf, sigma)
    return prf

def get_width(x1, y1, x2, y2):
    width_not_ordered = np.sqrt((x1-x2)**2+(y1-y2)**2)
    return width_not_ordered

def split_mesh2mesh(x1, y1, x2, y2):
    
    contour = mesh2contour(x1, y1, x2, y2)
    width = get_width(x1, y1, x2, y2)
    length = np.sum(get_step_length(x1, y1, x2, y2))
    
    x = (x1 + x2) / 2
    y = (y1 + y2) / 2
    line = np.array([x, y]).T
    midline, pole1, pole2 = extend_skeleton(line[3:-3], contour)
    l1, l2, profile_mesh, midline = straighten_by_orthogonal_lines(contour, midline, length, width, unit_micron= 0.5)
    result, l1, l2 = add_poles(l1, l2, pole1, pole2)
    return result, contour, midline

def get_profile_mesh(mesh, width, micron_unit = 1):
    l1 = mesh[:,:2]
    l2 = mesh[:,2:]
    sorted_width = sorted(width, reverse = True)
    width = (sum(sorted_width[:math.floor(len(sorted_width)/3)]) / math.floor(len(sorted_width)/3))
    N_width = int(round(np.median(width) / 0.05))
    dl = (l2 - l1) / N_width
    mult_mat = np.tile(np.arange(N_width + 1), (len(l1), 1))
    mat_x = l1[:, 0][:, np.newaxis] + mult_mat * dl[:, 0][:, np.newaxis]
    mat_y = l1[:, 1][:, np.newaxis] + mult_mat * dl[:, 1][:, np.newaxis]
    profile_mesh = np.array([mat_x, mat_y])
    return profile_mesh

def expand_contour(contour, scale= 1):
    """
    enlarges contour
    :param contour:
    :param scale:
    :return:
    """
    area = 0.5 * np.sum(np.diff(contour[:, 0]) * (contour[:-1, 1] + contour[1:, 1]))
    if area < 0:
        # If the area is negative, flip the sign of the unit perpendicular vector
        dxy = unit_perpendicular_vector(contour, closed=True) 
    else:
        # Otherwise, use the unit perpendicular vector as is
        dxy = unit_perpendicular_vector(contour, closed=True) * (-1)
    #dxy = unit_perpendicular_vector(contour, closed=True)

    return contour - (scale * dxy)

def erode_contour(contour, scale= 1):
    """
    shrinks contour
    :param contour:
    :param scale:
    :return:
    """
    area = 0.5 * np.sum(np.diff(contour[:, 0]) * (contour[:-1, 1] + contour[1:, 1]))
    if area < 0:
        # If the area is negative, flip the sign of the unit perpendicular vector
        dxy = unit_perpendicular_vector(contour, closed=True) 
    else:
        # Otherwise, use the unit perpendicular vector as is
        dxy = unit_perpendicular_vector(contour, closed=True) * (-1)
    #dxy = unit_perpendicular_vector(contour, closed=True)

    return contour + (scale * dxy)

def crop_image(contour, image, pad = 10):
    contour = contour.copy()
    contour = contour.astype(int)
    mask = polygon2mask(image.shape, contour)
    
    padding =10
    x, y, w, h = cv.boundingRect(contour)
    x = np.clip(x - padding, 0, image.shape[0])
    y = np.clip(y - padding, 0, image.shape[1])
    w = np.clip(w + 2 * padding, 0, image.shape[0] - x)
    h = np.clip(h + 2 * padding, 0, image.shape[1] - y)

    cropped_img = image[x:x+w, y:y+h]
    cropped_mask = mask[x:x+w, y:y+h]
    return cropped_img, cropped_mask, x, y



def get_object_contours(mask, x_offset, y_offset, smoothing=10):
    contours, _ = cv.findContours(mask.astype(np.uint8), cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)
    interpolated_contours = []

    for contour in contours:
        contour = np.array(contour)
        contourx = contour[:, 0, 1]
        contoury = contour[:, 0, 0]
        s = int(len(contourx) / smoothing) if len(contourx) > smoothing else 1
        tck, u = interpolate.splprep([contourx, contoury], u=None, s=s, per=1)
        u_new = np.linspace(u.min(), u.max(), len(contourx))
        outx, outy = interpolate.splev(u_new, tck, der=0)
        
        interpolated_contour = np.array([outx+x_offset, outy+y_offset]).T
        interpolated_contours.append(interpolated_contour)

    return interpolated_contours

def keep_significant_masks(nucleoid_mask, cropped_mask, min_overlap_ratio=0.01, max_external_ratio=0.1):
    # Find connected components (blobs) in the nucleoid_mask
    num_labels, labels, stats, _ = cv.connectedComponentsWithStats(nucleoid_mask.astype(np.uint8))

    # Create a mask to keep blobs with significant overlap
    significant_mask = np.zeros_like(nucleoid_mask, dtype=np.uint8)

    # Get the total area of the cropped_mask
    total_cropped_mask_area = np.sum(cropped_mask)

    for label in range(1, num_labels):  # Skip label 0 as it represents the background
        # Extract the blob region for the current label
        blob_mask = (labels == label).astype(np.uint8)

        # Calculate the overlap with the cropped_mask
        overlap_area = np.sum(np.logical_and(blob_mask, cropped_mask))

        # Calculate the overlap ratio
        overlap_ratio = overlap_area / total_cropped_mask_area

        # Calculate the external ratio (portion outside of cropped_mask)
        external_area = np.sum(np.logical_and(blob_mask, ~cropped_mask))
        external_ratio = external_area / np.sum(blob_mask)
        # Keep blobs that have a significant overlap with the cropped_mask
        # and don't exceed the maximum external ratio
        if overlap_ratio >= min_overlap_ratio and external_ratio <= max_external_ratio:
            significant_mask |= blob_mask

    return significant_mask


def get_subcellular_objects(contour, signal, cell_id, 
                            log_sigma, 
                            kernel_width, 
                            min_overlap_ratio=0.01, 
                            max_external_ratio=0.1):
    
    cropped_img, cropped_mask, x_offset, y_offset = crop_image(contour, signal)
    bgr = phase_background(cropped_img, cropped_mask)
    cropped_img_bgr = np.maximum(cropped_img - bgr, 0)

    # Apply Laplacian of Gaussian (LoG) filter to the cropped_img
    log_filtered_img = scipy.ndimage.gaussian_laplace(cropped_img_bgr, sigma=log_sigma)

    threshold_value = threshold_otsu(log_filtered_img)
    nucleoid_mask = log_filtered_img < threshold_value
    kernel = np.ones((kernel_width, kernel_width), np.uint8)  # You can adjust the kernel size as needed
    nucleoid_mask = cv.dilate(nucleoid_mask.astype(np.uint8), kernel, iterations=1)

    nucleoid_mask = keep_significant_masks(nucleoid_mask, cropped_mask, min_overlap_ratio, max_external_ratio)

    if np.any(nucleoid_mask == 1):
        object_contours = get_object_contours(nucleoid_mask, x_offset, y_offset)

    else:
        object_contours = None
        print(f'No nucleoids detected in cell {cell_id}')
    return object_contours








# -*- coding: utf-8 -*-
"""
Created on Mon Jul 31 16:46:21 2023

@author: Bart Steemans. Govers Lab.
"""
import pickle
import pandas as pd
import numpy as np

class Curation:
    
    def __init__(self, svm_df):
        
        self.svm_df = svm_df
        self.features = None
        self.curated_df = None
        self.svm_model = None
    
    def compiled_curation(self, path_to_model, cols):
        self.load_model(path_to_model)
        self.prepare_dataframe(cols)
        curated_df = self.make_predictions()
        return curated_df
        
    def load_model(self, path_to_model):
        with open(path_to_model, 'rb') as f:
            self.svm_model = pickle.load(f)
    
    def prepare_dataframe(self, cols):

        self.features = self.svm_df.iloc[:, cols:]
        
        self.features[self.features.select_dtypes(include=[np.number]).columns] = \
            self.features.select_dtypes(include=[np.number]).astype(float)
    
        # Check for infinite values and NaNs
        has_infinite = np.any(np.isinf(self.features.select_dtypes(include=[np.number])))
        has_nan = np.any(np.isnan(self.features.select_dtypes(include=[np.number])))
    
        if has_infinite or has_nan:
            # Create a boolean mask for rows without infinite or NaN values
            mask = ~(np.isinf(self.features.select_dtypes(include=[np.number])).any(axis=1) |
                     np.isnan(self.features.select_dtypes(include=[np.number])).any(axis=1))
    
            # Filter the DataFrame to keep only rows without infinite or NaN values
            svm_training_filtered = self.features[mask]
            
            # Print the number of removed rows
            num_removed_rows = len(self.features) - len(svm_training_filtered)
            print(f"\nNumber of removed rows: {num_removed_rows}")
    
            # Update svm_training with the filtered DataFrame
            self.features = svm_training_filtered
            self.svm_df = self.svm_df[mask]
            
        else:    
            print("\nNo rows removed")         
        
    def get_label_proportions(self):
        label_counts = self.curated_df['label'].value_counts()
        total_samples = len(self.curated_df)
        
        proportion_label_1 = label_counts.get(1, 0) / total_samples
        proportion_label_0 = label_counts.get(0, 0) / total_samples
        
        return proportion_label_1, proportion_label_0
    
    def make_predictions(self):
        # Make predictions using the SVM model
        predictions = self.svm_model.predict(self.features)
    
        # Assuming 'label' column is missing in the DataFrame, add the predicted labels
        self.curated_df = self.svm_df.copy()
        self.curated_df['label'] = predictions
    
        return self.curated_df
    
    def get_control(self, num_pos=5, num_neg=5):
        # Get the cell IDs of positive and negative classes from curated_df
        positive_cell_ids = self.curated_df[self.curated_df['label'] == 1]['cell_id'].tolist()
        negative_cell_ids = self.curated_df[self.curated_df['label'] == 0]['cell_id'].tolist()
        positive_cell_ids_random = np.random.choice(positive_cell_ids, size=num_pos, replace=False)
        negative_cell_ids_random = np.random.choice(negative_cell_ids, size=num_neg, replace=False)
    
        return np.array(positive_cell_ids_random, dtype = int), np.array(negative_cell_ids_random, dtype = int)